# -*- coding: utf-8 -*-
# pylint: disable=all
"""
Least Cost Xmission Command Line Interface
"""
import time
import click
import logging
import os
import json
from pathlib import Path
from warnings import warn

import numpy as np
import pandas as pd
import geopandas as gpd
from pyproj.crs import CRS
from shapely.geometry import Point

from rex.utilities.loggers import init_logger, init_mult, create_dirs
from rex.utilities.cli_dtypes import STR, INT
from rex.utilities.hpc import SLURM
from rex.utilities.utilities import get_class_properties, dict_str_load

from reV.handlers.exclusions import ExclusionLayers

from reVX.handlers.geotiff import Geotiff
from reVX.config.least_cost_xmission import LeastCostPathsConfig
from reVX.least_cost_xmission.config import XmissionConfig
from reVX.least_cost_xmission.least_cost_paths import (LeastCostPaths,
                                                       ReinforcementPaths)
from reVX.least_cost_xmission.least_cost_xmission import region_mapper
from reVX.least_cost_xmission.config.constants import (CELL_SIZE,
                                                       TRANS_LINE_CAT,
                                                       SUBSTATION_CAT,
                                                       BARRIERS_MULT,
                                                       BARRIER_H5_LAYER_NAME)
from reVX import __version__

logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('--name', '-n', default='LeastCostPaths', type=STR,
              show_default=True,
              help='Job name.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, name, verbose):
    """
    Least Cost Paths Command Line Interface
    """
    ctx.ensure_object(dict)
    ctx.obj['VERBOSE'] = verbose
    ctx.obj['NAME'] = name


@main.command()
def valid_config_keys():
    """
    Echo the valid Least Cost Paths config keys
    """
    click.echo(', '.join(get_class_properties(LeastCostPathsConfig)))


def run_local(ctx, config):
    """
    Run Least Cost Paths locally using config

    Parameters
    ----------
    ctx : click.ctx
        click ctx object
    config : reVX.config.least_cost_xmission.LeastCostPathsConfig
        Least Cost Paths config object.
    """
    ctx.obj['NAME'] = config.name
    ctx.invoke(local,
               cost_fpath=config.cost_fpath,
               features_fpath=config.features_fpath,
               cost_layers=config.cost_layers,
               network_nodes_fpath=config.network_nodes_fpath,
               transmission_lines_fpath=config.transmission_lines_fpath,
               xmission_config=config.xmission_config,
               capacity_class=config.capacity_class,
               clip_buffer=config.clip_buffer,
               start_index=0, step_index=1,
               tb_layer_name=config.tb_layer_name,
               barrier_mult=config.barrier_mult,
               max_workers=config.execution_control.max_workers,
               region_identifier_column=config.region_identifier_column,
               save_paths=config.save_paths,
               out_dir=config.dirout,
               log_dir=config.log_directory,
               ss_id_col=config.ss_id_col,
               verbose=config.log_level,
               li_cost_layers=config.length_invariant_cost_layers,
               tracked_layers=config.tracked_layers,
               cell_size=config.cell_size)


@main.command()
@click.option('--config', '-c', required=True,
              type=click.Path(exists=True),
              help='Filepath to AssemblyAreas config json file.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def from_config(ctx, config, verbose):
    """
    Run Least Cost Paths from a config.
    """

    config = LeastCostPathsConfig(config)

    if 'VERBOSE' in ctx.obj:
        if any((ctx.obj['VERBOSE'], verbose)):
            config._log_level = logging.DEBUG
    elif verbose:
        config._log_level = logging.DEBUG

    if config.execution_control.option == 'local':
        run_local(ctx, config)
        return

    if config.execution_control.option not in {'eagle', 'kestrel'}:
        click.echo('Option "{}" is not supported'
                   .format(config.execution_control.option))
        return

    # No need to add index to file name
    if config.execution_control.nodes == 1:
        eagle(config)
        return

    name = config.name
    num_nodes = config.execution_control.nodes
    logger.info('Splitting features over {} SLURM jobs'.format(num_nodes))
    n_zfill = len(str(num_nodes))
    for i in range(num_nodes):
        config.name = '{}_{}'.format(name, str(i).zfill(n_zfill))
        eagle(config, start_index=i)


@main.command()
@click.option('--cost_fpath', '-cost', type=click.Path(exists=True),
              required=True,
              help=("Path to h5 file with cost rasters and other required "
                    "layers"))
@click.option('--features_fpath', '-feats', required=True,
              type=click.Path(exists=True),
              help="Path to GeoPackage with transmission features")
@click.option('--cost-layers', '-cl', required=True, multiple=True,
              default=(),
              help='Layer in H5 to add to total cost raster used for routing. '
                   'Multiple layers may be specified. If running '
                   'reinforcement computations, layer name may have '
                   'curly brackets (``{}``), which will be filled in '
                   'based on the capacity class input (e.g. '
                   '"tie_line_costs_{}MW")')
@click.option('--network_nodes_fpath', '-nn', type=STR, show_default=True,
              default=None,
              help=("Path to Network Nodes GeoPackage. If given alongside "
                    "`transmission_lines_fpath`, reinforcement path cost "
                    "calculation is run."))
@click.option('--transmission_lines_fpath', '-tl', type=STR, show_default=True,
              default=None,
              help=("Path to Transmission lines GeoPackage. This file can "
                    "contain other features, but transmission lines must "
                    "be identified by {!r}. If given alongside "
                    "`network_nodes_fpath`, reinforcement path cost "
                    "calculation is run.".format(TRANS_LINE_CAT)))
@click.option('--xmission_config', '-xcfg', type=STR, show_default=True,
              default=None,
              help=("Path to transmission config .json. Only used for "
                    "reinforcement path computations."))
@click.option('--capacity_class', '-cap', type=str, required=False,
              help=("Capacity class of the 'base' greenfield costs layer. "
                    "Costs will be scaled by the capacity corresponding to "
                    "this class to report reinforcement costs as $/MW. "
                    "Only used for reinforcement path computations."))
@click.option('--clip_buffer', '-cb', type=int,
              show_default=True, default=0,
              help="Optional number of array elements to buffer clip area by.")
@click.option('--start_index', '-start', type=int,
              show_default=True, default=0,
              help=("Start index of features to run."))
@click.option('--step_index', '-step', type=int,
              show_default=True, default=1,
              help=("Step index of features to run."))
@click.option('--tb_layer_name', '-tbln', default=BARRIER_H5_LAYER_NAME,
              type=STR, show_default=True,
              help='Name of transmission barrier layer in `cost_fpath` file. '
                   'This layer defines the multipliers applied to the cost '
                   'layer to determine LCP routing (but does not actually '
                   'affect output costs')
@click.option('--barrier_mult', '-bmult', type=float,
              show_default=True, default=BARRIERS_MULT,
              help=("Transmission barrier multiplier, used when computing the "
                    "least cost tie-line path"))
@click.option('--max_workers', '-mw', type=INT,
              show_default=True, default=None,
              help=("Number of workers to use for processing, if 1 run in "
                    "serial, if None use all available cores"))
@click.option('--region_identifier_column', '-rid', type=STR, default=None,
              help=("Name of column in reinforcement regions GeoPackage"
                    "containing a unique identifier for each region."))
@click.option('--save_paths', '-paths', is_flag=True,
              help="Flag to save least cost path as a multi-line geometry")
@click.option('--out_dir', '-o', type=STR, default='./',
              show_default=True,
              help='Directory to save least cost Paths values to.')
@click.option('--log_dir', '-log', default=None, type=STR,
              show_default=True,
              help='Directory to dump log files.')
@click.option('--ss_id_col', '-ssid', default="poi_gid", type=STR,
              show_default=True,
              help='Name of column used to unqiuely identify substations. '
                   'Used for reinforcement calcaultions only. ')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.option('--li-cost-layers', '-licl', required=False, multiple=True,
              default=(),
              help='Length-invariant cost layer in H5 to add to total cost '
                   'raster used for routing. These costs do not scale with '
                   'distance traversed acroiss the cell. Multiple layers may '
                   'be specified.')
@click.option('--tracked_layers', '-trl', type=STR, default=None,
              show_default=True,
              help=('Dictionary mapping layer names to strings, where the '
                    'strings are numpy methods that should be applied to '
                    'the layer along the LCP.'))
@click.option('--cell_size', '-cs', type=INT,
              show_default=True, default=CELL_SIZE,
              help=("Side length of a single cell in meters. Cells are "
                    "assumed to be square. Default is"))
@click.pass_context
def local(ctx, cost_fpath, features_fpath, cost_layers, network_nodes_fpath,
          transmission_lines_fpath, xmission_config, capacity_class,
          clip_buffer, start_index, step_index, tb_layer_name, barrier_mult,
          max_workers, region_identifier_column, save_paths, out_dir, log_dir,
          ss_id_col, verbose, li_cost_layers, tracked_layers, cell_size):
    """
    Run Least Cost Paths on local hardware
    """
    start_time = time.time()
    name = ctx.obj['NAME']
    if 'VERBOSE' in ctx.obj:
        verbose = any((ctx.obj['VERBOSE'], verbose))

    log_modules = [__name__, 'reVX', 'reV', 'rex']
    init_mult(name, log_dir, modules=log_modules, verbose=verbose)

    create_dirs(out_dir)
    logger.info('Computing Least Cost Paths connections and writing them to {}'
                .format(out_dir))

    logger.debug("Tracked layers input: %r", tracked_layers)
    if isinstance(tracked_layers, str):
        tracked_layers = dict_str_load(tracked_layers)
    cost_layers = list(cost_layers)
    li_cost_layers = list(li_cost_layers)

    is_reinforcement_run = (network_nodes_fpath is not None
                            and transmission_lines_fpath is not None)

    kwargs = {"clip_buffer": int(clip_buffer),
              "tb_layer_name": tb_layer_name,
              "barrier_mult": barrier_mult,
              "save_paths": save_paths,
              "max_workers": max_workers,
              "length_invariant_cost_layers": li_cost_layers,
              "tracked_layers": tracked_layers,
              "cell_size": cell_size}

    if is_reinforcement_run:
        logger.info('Detected reinforcement run!')
        xmission_config = XmissionConfig(config=xmission_config)
        kwargs["xmission_config"] = xmission_config
        cc_str = xmission_config._parse_cap_class(capacity_class)
        cap = xmission_config['power_classes'][cc_str]
        cost_layers = [layer.format(cap) for layer in cost_layers]
        logger.debug('Xmission Config: {}'.format(xmission_config))
        features = gpd.read_file(network_nodes_fpath)
        features, *__ = LeastCostPaths._map_to_costs(cost_fpath, features)
        kwargs["indices"] = features.index[start_index::step_index]
        kwargs["ss_id_col"] = ss_id_col
        least_costs = ReinforcementPaths.run(cost_fpath, features_fpath,
                                             network_nodes_fpath,
                                             region_identifier_column,
                                             transmission_lines_fpath,
                                             capacity_class, cost_layers,
                                             **kwargs)
    else:
        features = gpd.read_file(features_fpath)
        features, *__ = LeastCostPaths._map_to_costs(cost_fpath, features)
        kwargs["indices"] = features.index[start_index::step_index]
        least_costs = LeastCostPaths.run(cost_fpath, features_fpath,
                                         cost_layers, **kwargs)

    fpath_out = os.path.join(out_dir, f'{name}_lcp')
    if save_paths:
        fpath_out += '.gpkg'
        logger.info('Writing output to %s', fpath_out)
        least_costs.to_file(fpath_out, driver="GPKG", index=False)
    else:
        fpath_out += '.csv'
        logger.info('Writing output to %s', fpath_out)
        least_costs.to_csv(fpath_out, index=False)

    logger.info('Writing output complete')
    elapsed_time = (time.time() - start_time) / 60
    logger.info("Processing took %.2f minutes", elapsed_time)


@main.command()
@click.option('--features_fpath', '-feats', required=True,
              type=click.Path(exists=True),
              help="Path to GeoPackage with substation and transmission "
                   "features")
@click.option('--regions_fpath', '-regs', required=True,
              type=click.Path(exists=True),
              help=("Path to reinforcement regions GeoPackage."))
@click.option('--region_identifier_column', '-rid', required=True,
              type=STR,
              help=("Name of column in reinforcement regions GeoPackage"
                    "containing a unique identifier for each region."))
@click.option('--out_file', '-of', default=None, type=STR,
              help='Name for output GeoPackage file.')
@click.pass_context
def map_ss_to_rr(ctx, features_fpath, regions_fpath, region_identifier_column,
                 out_file):
    """
    Map substation locations to reinforcement regions.

    Reinforcement regions are user-defined. Typical regions are
    Balancing Areas, States, or Counties, though custom regions are also
    allowed. Each region must be supplied with a unique identifier in
    the input file.

    This method also removes substations that do not meet the min 69 kV
    voltage requirement and adds {'min_volts', 'max_volts'} fields to
    the remaining substations.

    .. Important:: This method DOES NOT clip the substations to the
      reinforcement regions boundary. All substations will be mapped to
      their closest region. It is your responsibility to remove any
      substations outside of the analysis region before calling this
      method.

    Doing the pre-processing step avoids any issues with substations
    being left out or double counted if they were simply clipped to the
    reinforcement region shapes.
    """
    log_level = "DEBUG" if ctx.obj.get('VERBOSE') else "INFO"
    init_logger('reVX', log_level=log_level)

    features = gpd.read_file(features_fpath)
    substations = (features[features.category == SUBSTATION_CAT]
                   .reset_index(drop=True)
                   .dropna(axis="columns", how="all"))

    regions = gpd.read_file(regions_fpath).to_crs(features.crs)
    logger.info("Mapping {:,d} substation locations to {:,d} "
                "reinforcement regions"
                .format(substations.shape[0], regions.shape[0]))

    map_func = region_mapper(regions, region_identifier_column)
    centroids = substations.centroid
    substations[region_identifier_column] = centroids.apply(map_func)

    logger.info("Calculating min/max voltage for each substation...")
    bad_subs = np.zeros(len(substations), dtype=bool)
    for idx, row in substations.iterrows():
        lines = row['trans_gids']
        if isinstance(lines, str):
            lines = json.loads(lines)

        lines_mask = features['gid'].isin(lines)
        voltage = features.loc[lines_mask, 'voltage'].values

        if np.max(voltage) >= 69:
            substations.loc[idx, 'min_volts'] = np.min(voltage)
            substations.loc[idx, 'max_volts'] = np.max(voltage)
        else:
            bad_subs[idx] = True

    if any(bad_subs):
        msg = ("The following sub-stations do not have the minimum "
               "required voltage of 69 kV and will be dropped:\n{}"
               .format(substations.loc[bad_subs, 'gid']))
        logger.warning(msg)
        warn(msg)
        substations = substations.loc[~bad_subs].reset_index(drop=True)

    logger.info("Writing substation output to {!r}".format(out_file))
    substations.to_file(out_file, driver="GPKG", index=False)


@main.command()
@click.option('--connections_fpath', '-con', required=True,
              type=click.Path(exists=True),
              help="Path to GeoPackage with substation and transmission "
                   "features")
@click.option('--region_identifier_column', '-rid', required=True,
              type=STR,
              help=("Name of column in reinforcement regions GeoPackage"
                    "containing a unique identifier for each region."))
@click.option('--out_file', '-of', required=True, type=STR,
              help='Name for output GeoPackage file.')
@click.pass_context
def ss_from_conn(ctx, connections_fpath, region_identifier_column, out_file):
    """Extract substations from connections table output by LCP.

    Substations extracted by this method can be usef for reinforcement
    calculations.

    This method also does minor filtering/formatting of the input
    connections table.

    Reinforcement regions are user-defined. Typical regions are
    Balancing Areas, States, or Counties, though custom regions are also
    allowed. Each region must be supplied with a unique identifier in
    the input file.

    .. Important:: This method DOES NOT clip the substations to the
      reinforcement regions boundary. All substations will be mapped to
      their closest region. It is your responsibility to remove any
      substations outside of the analysis region before calling this
      method.

    Doing the pre-processing step avoids any issues with substations
    being left out or double counted if they were simply clipped to the
    reinforcement region shapes.
    """
    log_level = "DEBUG" if ctx.obj.get('VERBOSE') else "INFO"
    init_logger('reVX', log_level=log_level)

    logger.info("Reading in connection info...")
    if connections_fpath.endswith(".csv"):
        connections = pd.read_csv(connections_fpath)
    elif connections_fpath.endswith(".gpkg"):
        connections = gpd.read_file(connections_fpath)
    else:
        raise ValueError("Unknown file ending for features file (must be "
                         f"'.csv' or '.gpkg'): {connections_fpath}")

    logger.info("Filtering out NaN's in connection info...")
    connections = connections[~connections["poi_gid"].isna()]
    connections = connections[~connections["poi_lat"].isna()]
    connections = connections[~connections["poi_lon"].isna()]

    logger.info("Extracting substation locations...")
    cols = ['poi_gid', 'poi_lat', 'poi_lon', region_identifier_column]
    poi_groups = connections.groupby(cols)
    pois = [info for info, __ in poi_groups]
    pois = pd.DataFrame(pois, columns=cols)

    geo = [Point(row["poi_lon"], row["poi_lat"])
           for __, row in pois.iterrows()]
    substations = gpd.GeoDataFrame(pois, crs="epsg:4326", geometry=geo)

    logger.info("Writing substation output to {!r}".format(out_file))
    substations["poi_gid"] = substations["poi_gid"].astype("Int64")
    substations.to_file(out_file, driver="GPKG", index=False)


@main.command()
@click.option('--network_nodes_fpath', '-nodes', type=click.Path(exists=True),
              help=("Path to network nodes GeoPackage. The "
                    "`region_identifier_column` will be added to this file if "
                    "it is missing."))
@click.option('--regions_fpath', '-regs', required=True,
              type=click.Path(exists=True),
              help=("Path to reinforcement regions GeoPackage. The "
                    "`region_identifier_column` from this file will be "
                    "added to the `network_nodes_fpath` if needed."))
@click.option('--region_identifier_column', '-rid', required=True,
              type=STR,
              help=("Name of column in reinforcement regions GeoPackage"
                    "containing a unique identifier for each region."))
@click.option('--crs_tempalate_file', '-ctf', default=None, type=STR,
              help=("Path to a fiel containing the CRS that should be used "
                    "to perform the mapping. This input can be an exclusions "
                    "HDF5 file, a GeoTIFF, or any file tha can be read with "
                    "GeoPandas. If `None`, the `network_nodes_fpath` file "
                    "CRS is used to perfom the mapping."))
@click.option('--out_file', '-of', default=None, type=STR,
              help='Name for output GeoPackage file. If `None`, the '
                   '`network_nodes_fpath` file will be overwritten. ')
@click.pass_context
def add_rr_to_nn(ctx, network_nodes_fpath, regions_fpath,
                 region_identifier_column, crs_tempalate_file, out_file):
    """Add reinforcement region column to network node file.

    Reinforcement regions are user-defined. Typical regions are
    Balancing Areas, States, or Counties, though custom regions are also
    allowed. Each region must be supplied with a unique identifier in
    the input file.
    """
    log_level = "DEBUG" if ctx.obj.get('VERBOSE') else "INFO"
    init_logger('reVX', log_level=log_level)
    crs_tempalate_file = Path(crs_tempalate_file or network_nodes_fpath)

    logger.info("Reading in network node info...")
    network_nodes_fpath = Path(network_nodes_fpath)
    network_nodes = gpd.read_file(network_nodes_fpath)

    logger.info("Reading in CRS template...")
    if crs_tempalate_file.suffix == ".h5":
        with ExclusionLayers(str(crs_tempalate_file)) as excl:
            crs = CRS.from_string(excl.crs)
    elif crs_tempalate_file.suffix in {".tif", ".tiff"}:
        with Geotiff(str(crs_tempalate_file)) as geo:
            crs = geo.profile["crs"]
    else:
        crs = gpd.read_file(crs_tempalate_file).crs

    network_nodes = network_nodes.to_crs(crs)
    regions = gpd.read_file(regions_fpath).to_crs(crs)
    if region_identifier_column in network_nodes:
        msg = ("Network nodes file {!r} was specified but it "
               "already contains the {!r} column. No data modified!"
               .format(str(network_nodes_fpath), region_identifier_column))
        logger.warning(msg)
        warn(msg)
        return

    logger.info("Adding regions to network nodes...")
    centroids = network_nodes.centroid
    map_func = region_mapper(regions, region_identifier_column)
    network_nodes[region_identifier_column] = centroids.apply(map_func)

    if out_file is None:
        out_fn = "{}.gpkg".format(network_nodes_fpath.stem)
        out_file = network_nodes_fpath.parent / out_fn

    logger.info("Writing updated network node data to {!r}"
                .format(str(out_file)))
    network_nodes.to_file(out_file, driver="GPKG")


def get_node_cmd(config, start_index=0):
    """
    Get the node CLI call for Least Cost Paths

    Parameters
    ----------
    config : reVX.config.least_cost_xmission.LeastCostPathsConfig
        Least Cost Paths config object.

    Returns
    -------
    cmd : str
        CLI call to submit to SLURM execution.
    """

    args = ['-n {}'.format(SLURM.s(config.name)),
            'local',
            '-cost {}'.format(SLURM.s(config.cost_fpath)),
            '-feats {}'.format(SLURM.s(config.features_fpath)),
            '-nn {}'.format(SLURM.s(config.network_nodes_fpath)),
            '-tl {}'.format(SLURM.s(config.transmission_lines_fpath)),
            '-rid {}'.format(SLURM.s(config.region_identifier_column)),
            '-cap {}'.format(SLURM.s(config.capacity_class)),
            '-cb {}'.format(SLURM.s(config.clip_buffer)),
            '-start {}'.format(SLURM.s(start_index)),
            '-step {}'.format(SLURM.s(config.execution_control.nodes or 1)),
            '-tbln {}'.format(SLURM.s(config.tb_layer_name)),
            '-bmult {}'.format(SLURM.s(config.barrier_mult)),
            '-ssid {}'.format(SLURM.s(config.ss_id_col)),
            '-mw {}'.format(SLURM.s(config.execution_control.max_workers)),
            '-o {}'.format(SLURM.s(config.dirout)),
            '-log {}'.format(SLURM.s(config.log_directory)),
            '-cs {}'.format(SLURM.s(config.cell_size)),
            ]

    for layer in config.cost_layers:
        args.append(f'-cl {layer}')
    for layer in config.length_invariant_cost_layers:
        args.append(f'-licl {layer}')

    if config.save_paths:
        args.append('-paths')

    if config.tracked_layers:
        args.append('-trl {}'.format(SLURM.s(config.tracked_layers)))

    if config.log_level == logging.DEBUG:
        args.append('-v')

    if config.xmission_config is not None:
        args.append('-xcfg {}'.format(SLURM.s(config.xmission_config)),)

    cmd = ('python -m reVX.least_cost_xmission.least_cost_paths_cli {}'
           .format(' '.join(args)))
    logger.debug('Submitting the following cli call:\n\t{}'.format(cmd))

    return cmd


def eagle(config, start_index=0):
    """
    Run Least Cost Paths on Eagle HPC.

    Parameters
    ----------
    config : reVX.config.least_cost_xmission.LeastCostPathsConfig
        Least Cost Paths config object.
    """

    cmd = get_node_cmd(config, start_index)
    name = config.name
    log_dir = config.log_directory
    stdout_path = os.path.join(log_dir, 'stdout/')

    slurm_manager = SLURM()

    logger.info('Running Least Cost Paths on Eagle with '
                'node name "{}"'.format(name))
    out = slurm_manager.sbatch(cmd,
                               alloc=config.execution_control.allocation,
                               memory=config.execution_control.memory,
                               walltime=config.execution_control.walltime,
                               feature=config.execution_control.feature,
                               name=name, stdout_path=stdout_path,
                               conda_env=config.execution_control.conda_env,
                               module=config.execution_control.module)[0]
    if out:
        msg = ('Kicked off Least Cost Paths "{}" '
               '(SLURM jobid #{}) on Eagle.'
               .format(name, out))
    else:
        msg = ('Was unable to kick off Least Cost Paths '
               '"{}". Please see the stdout error messages'
               .format(name))

    click.echo(msg)
    logger.info(msg)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running Least Cost Paths CLI')
        raise



"""
CLI to create wet and dry costs, barriers, and friction layers. Final
layers required for LCP routing are saved to an H5 file. All layers may
optionally be saved to GeoTIFF.
"""
import sys
import click
import logging
from pathlib import Path
from warnings import warn

import numpy as np
from pydantic import ValidationError
from rex.utilities.loggers import init_mult
from gaps.config import load_config

from reVX import __version__
from reVX.config.transmission_layer_creation import (
    TransmissionLayerCreationConfig, MergeFrictionBarriers
)
from reVX.least_cost_xmission.config.constants import ALL
from reVX.handlers.layered_h5 import LayeredTransmissionH5
from reVX.least_cost_xmission.layers import LayerCreator
from reVX.least_cost_xmission.layers.masks import Masks
from reVX.least_cost_xmission.layers.utils import convert_pois_to_lines
from reVX.least_cost_xmission.layers.dry_cost_creator import DryCostCreator


logger = logging.getLogger(__name__)
CONFIG_ACTIONS = ['layers', 'dry_costs', 'merge_friction_and_barriers']


@click.group()
@click.version_option(version=__version__)
@click.option('--verbose', '-v', is_flag=True, default=False,
              help='Flag to turn on debug logging. Default is not verbose.')
def main(verbose):
    """
    Transmission LCP routing cost, friction, and barrier layer creator.
    """
    log_modules = [__name__, 'reVX', 'reV', 'rex']
    init_mult('trans-creator', None, modules=log_modules, verbose=verbose)


@main.command
@click.option('-c', '--config', 'config_fpath', type=click.Path(exists=True),
              required=True, help='Configuration JSON.')
def from_config(config_fpath: str):  # noqa: C901
    """
    Create costs, barriers, and frictions from a config file.
    """
    config_dict = load_config(config_fpath)
    try:
        config = TransmissionLayerCreationConfig.model_validate(config_dict)
    except ValidationError as e:
        logger.error(f'Error loading config file {config_fpath}:\n{e}')
        sys.exit(1)

    if not any(map(lambda key: config.model_dump()[key] is not None,
                   CONFIG_ACTIONS)):
        logger.error(f'At least one of {CONFIG_ACTIONS} must be in the '
                     'config file')
        sys.exit(1)

    # Done with guard clauses
    output_tiff_dir = Path(config.output_tiff_dir).expanduser().resolve()
    output_tiff_dir.mkdir(exist_ok=True, parents=True)
    template_file = str(config.template_raster_fpath)
    h5_io_handler = LayeredTransmissionH5(h5_file=str(config.h5_fpath),
                                          template_file=template_file,
                                          layer_dir=config.layer_dir)

    masks = _load_masks(config, h5_io_handler)

    # Perform actions in config
    builder = LayerCreator(h5_io_handler, masks, output_tiff_dir,
                           cell_size=config.cell_size)
    for lc in config.layers or []:
        builder.build(lc.layer_name, lc.build,
                      values_are_costs_per_mile=lc.values_are_costs_per_mile,
                      write_to_h5=lc.include_in_h5, description=lc.description)

    if config.dry_costs is not None:
        dc = config.dry_costs
        template_file = str(dc.iso_region_tiff)

        try:
            dry_mask = masks.dry_mask
        except ValueError:
            msg = "Dry mask not found! Computing dry costs for full extent!"
            logger.warning(msg)
            warn(msg)
            dry_mask = np.full(h5_io_handler.shape, True)

        dcc = DryCostCreator(h5_io_handler, dry_mask, output_tiff_dir,
                             cell_size=config.cell_size)
        cost_configs = None if not dc.cost_configs else str(dc.cost_configs)
        dcc.build(str(dc.iso_region_tiff), str(dc.nlcd_tiff),
                  str(dc.slope_tiff), cost_configs=cost_configs,
                  default_mults=dc.default_mults, extra_tiffs=dc.extra_tiffs)

    if config.merge_friction_and_barriers is not None:
        _combine_friction_and_barriers(config.merge_friction_and_barriers,
                                       h5_io_handler, output_tiff_dir)


@main.command
@click.option('--poi-file', '-p', required=True, type=click.Path(exists=True),
              help='File of POIs in CSV format. Each POI must have the '
              'following fields: "POI Name", "State", "Voltage (kV)", "Lat", '
              'and "Long". "State" may be blank. Other fields are ignored.')
@click.option('--template-raster', '-t', required=True,
              type=click.Path(exists=True),
              help='Raster to extract CRS from.')
@click.option('--out-file', '-o', type=click.Path(), required=True,
              help='Filename to use for POI lines GeoPackage file.')
def convert_pois(poi_file: str, template_raster: str, out_file: str):
    """
    Convert points of interconnection (POI) to short lines. The transmission
    routing code requires all transmission elements to be lines. The POIs
    defined in the CSV will be converted to lines and labeled as substations.
    As all substations must be link to a transmission line, a synthetic
    transmission line is created that is linked to the POIs.
    """
    convert_pois_to_lines(poi_file, template_raster, out_file)


@main.command
@click.option('--land-mask-vector', '-l', required=True,
              type=click.Path(exists=True),
              help='GeoPackage or shapefile representing land.')
@click.option('--template-raster', '-t', required=True,
              type=click.Path(exists=True),
              help='Raster to extract CRS, transform, and shape from.')
@click.option('--masks-dir', '-m', type=click.Path(), required=False,
              default='.',
              help='Directory to store mask GeoTIFFs in.')
@click.option('--dont-reproject', '-d', is_flag=True, default=False,
              help='Don\'t reproject vector before creating masks.')
def create_masks(land_mask_vector: str, template_raster: str, masks_dir: str,
                 dont_reproject: bool):
    """
    Convert a vector land file to wet, dry, and all other masks.
    """
    io_handler = LayeredTransmissionH5(template_file=template_raster)
    masks = Masks(io_handler, masks_dir=masks_dir)

    reproject = not dont_reproject
    masks.create_masks(land_mask_vector, save_tiff=True,
                       reproject_vector=reproject)


@main.command
@click.option('--template-raster', '-t', type=click.Path(exists=True),
              required=True,
              help='Raster to extract CRS, transform, and shape from.')
@click.option('--h5-file', '-h', type=click.Path(exists=False), required=True,
              help='Name of H5 file to create.')
@click.option('--block_size', '-bs', type=int, required=False, default=None,
              help='Block size used to build lat/lon datasets.')
def create_h5(template_raster: str, h5_file: str, block_size: int):
    """
    Create a new H5 file to store layers in.
    """
    logger.info('Using raster %s to create new H5 file %s', template_raster,
                h5_file)
    lth5 = LayeredTransmissionH5(h5_file, template_file=template_raster,
                                 block_size=block_size)
    lth5.create_new()


def _load_masks(config, h5_io_handler):
    """Load masks based on config file.

    Parameters
    ----------
    config : :class:`MergeFrictionBarriers`
        Config object
    h5_io_handler : :class:`LayeredTransmissionH5`
        Transmission IO handler

    Returns
    -------
    Masks
        Masks instance based on directories in config file.
    """
    masks = Masks(h5_io_handler, masks_dir=config.masks_dir)
    if not config.layers:
        return masks

    build_configs = [lc.build for lc in config.layers]
    need_masks = any(lc.extent != ALL
                     for bc in build_configs for lc in bc.values())
    if need_masks:
        masks.load_masks()

    return masks


def _combine_friction_and_barriers(config: MergeFrictionBarriers,
                                   io_handler: LayeredTransmissionH5,
                                   output_tiff_dir=None):
    """
    Combine friction and barriers and save to H5 and optionally GeoTIFF

    Parameters
    ----------
    config : :class:`MergeFrictionBarriers`
        Config object
    io_handler : :class:`LayeredTransmissionH5`
        Transmission IO handler
    output_tiff_dir : path-like, optional
        Directory where combined barriers should be saved as GeoTIFF. If
        ``None``, combined layers are not saved. By default, ``None``.
    """
    output_tiff_dir = Path(output_tiff_dir)
    friction_tiff = output_tiff_dir / f"{config.friction_layer}.tif"
    raw_barrier_tiff = output_tiff_dir / f"{config.barrier_layer}.tif"
    if not friction_tiff.exists():
        logger.error(f'The friction GeoTIFF ({str(friction_tiff)}) was not '
                     'found. Please create it using the `friction_layers` '
                     'key in the config file.')
        sys.exit(1)

    if not raw_barrier_tiff.exists():
        logger.error(f'The raw barriers GeoTIFF ({str(raw_barrier_tiff)}) was '
                     'not found. Please create it using the `barrier_layers` '
                     'key in the config file.')
        sys.exit(1)

    logger.info('Loading friction and raw barriers.')
    friction = io_handler.load_data_using_h5_profile(friction_tiff)
    barriers = io_handler.load_data_using_h5_profile(raw_barrier_tiff)

    combined = friction + barriers * config.barrier_multiplier

    if output_tiff_dir is not None:
        out_fp = output_tiff_dir / f"{config.output_layer_name}.tif"
        logger.debug('Saving combined barriers to %s', out_fp)
        io_handler.save_data_using_h5_profile(combined, out_fp)

    logger.info('Writing combined barriers to H5')
    io_handler.write_layer_to_h5(combined, config.output_layer_name)


if __name__ == '__main__':
    try:
        main()  # pylint: disable=no-value-for-parameter
    except Exception:
        logger.exception('Error running Offshore Layer Creator CLI')
        raise



# -*- coding: utf-8 -*-
"""
Module to compute least cost transmission paths, distances, and costs
for a clipped area.
"""
from itertools import chain
import geopandas as gpd
import logging
import numpy as np
import numpy.typing as npt
import pandas as pd
import rasterio
from shapely.geometry import Polygon, Point
from shapely.geometry.linestring import LineString
from shapely.ops import nearest_points
from skimage.graph import MCP_Geometric
import time
from warnings import warn
from typing import Union, Optional, Tuple, Dict

from reV.handlers.exclusions import ExclusionLayers

from reVX.least_cost_xmission.config import XmissionConfig
from reVX.least_cost_xmission.config.constants import (TRANS_LINE_CAT,
                                                       SINK_CAT,
                                                       SINK_CONNECTION_COST,
                                                       SUBSTATION_CAT,
                                                       LOAD_CENTER_CAT,
                                                       CELL_SIZE,
                                                       SHORT_MULT,
                                                       MEDIUM_MULT,
                                                       SHORT_CUTOFF,
                                                       MEDIUM_CUTOFF,
                                                       MINIMUM_SPUR_DIST_KM,
                                                       BARRIERS_MULT,
                                                       BARRIER_H5_LAYER_NAME,
                                                       ISO_H5_LAYER_NAME)
from reVX.utilities.exceptions import (InvalidMCPStartValueError,
                                       LeastCostPathNotFoundError)

logger = logging.getLogger(__name__)


class TieLineCosts:
    """
    Compute Least Cost Tie-line cost from start location to desired end
    locations
    """

    def __init__(self, cost_fpath, start_indices, cost_layers,
                 row_slice, col_slice, xmission_config=None,
                 tb_layer_name=BARRIER_H5_LAYER_NAME,
                 barrier_mult=BARRIERS_MULT,
                 length_invariant_cost_layers=None, tracked_layers=None,
                 cell_size=CELL_SIZE):
        """
        Parameters
        ----------
        cost_fpath : str
            Full path to HDF5 file containing cost arrays. The cost data
            layers should be named ``"tie_line_costs_{capacity}MW"``,
            where ``capacity`` is an integer value representing the
            capacity of the line (the integer values must matches at
            least one of the integer capacity values represented by the
            keys in the ``power_classes`` portion of the transmission
            config).
        start_indices : tuple
            Tuple of (row_idx, col_idx) in the **clipped** cost array
            (see `row_slice` and `col_slice` definitions below)
            indicating the start position of all paths to compute
            (typically, this is the centroid of the supply curve cell).
            Paths will be computed from this start location to each of
            the ``end_indices``, which are also locations in the cost
            array (typically transmission feature locations).
        cost_layers : List[str]
            List of layers in H5 that are summed to determine total
            costs raster used for routing. Costs and distances for each
            individual layer are also reported (e.g. wet and dry costs).
            deteremining path using main cost layer.
        row_slice, col_slice : slice
            Slices into the cost raster array used to clip the area that
            should be considered when computing a least cost path. This
            can be used to limit the consideration space and speed up
            computation. Note that the start and end indices must
            be given w.r.t. the cost raster that is "clipped" using
            these slices.
        xmission_config : str | dict | XmissionConfig, optional
            Path to transmission config JSON files, dictionary of
            transmission config JSONs, or preloaded XmissionConfig
            objects. If ``None``, the default config is used.
            By default, ``None``.
        tb_layer_name : str, default=:obj:`BARRIER_H5_LAYER_NAME`
            Name of transmission barrier layer in `cost_fpath` file.
            This layer defines the multipliers applied to the cost layer
            to determine LCP routing (but does not actually affect
            output costs). By default, :obj:`BARRIER_H5_LAYER_NAME`.
        barrier_mult : int, optional
            Multiplier applied to the cost data to create transmission
            barrier costs. By default, ``100``.
        length_invariant_cost_layers : List[str] | None, optional
            List of layers in H5 to be added to the cost raster. The
            costs specified by these layers are not scaled with distance
            traversed across the cell (i.e. fixed one-time costs for
            crossing these cells).
        tracked_layers : dict, optional
            Dictionary mapping layer names to strings, where the strings
            are numpy methods that should be applied to the layer along
            the LCP. For example,
            ``tracked_layers={'layer_1': 'mean', 'layer_2': 'max}``
            would report the average of ``layer_1`` values along the
            least cost path and the max of ``layer_2`` values along the
            least cost path. Examples of numpy methods (non-exhaustive):

                - mean
                - max
                - min
                - mode
                - median
                - std

            By default, ``None``, which does not track any extra layers.
        cell_size : int, optional
            Side length of each cell, in meters. Cells are assumed to be
            square. By default, :obj:`CELL_SIZE`.
        """
        self._cost_fpath = cost_fpath
        self._tb_layer_name = tb_layer_name
        self._config = self._parse_config(xmission_config=xmission_config)
        self._start_indices = start_indices
        self._row_slice = row_slice
        self._col_slice = col_slice
        self._cell_size = cell_size
        self._clip_shape = self._mcp = self._cost = self._mcp_cost = None
        self._cost_layer_map = {}
        self._li_cost_layer_map = {}
        self._tracked_layers = tracked_layers or {}
        self._tracked_layer_map = {}
        self._cumulative_costs = None
        self.transform = None
        self._full_shape = None
        self._cost_crs = None

        with ExclusionLayers(self._cost_fpath) as fh:
            self._extract_data_from_cost_h5(fh)

        licl = length_invariant_cost_layers
        self._clip_costs(cost_layers, length_invariant_cost_layers=licl,
                         barrier_mult=barrier_mult)

        self._null_extras = {}
        for layer in chain(self._cost_layer_map, self._li_cost_layer_map):
            self._null_extras[f'{layer}_cost'] = np.nan
            self._null_extras[f'{layer}_dist_km'] = np.nan
        for layer_name in self._tracked_layer_map:
            method = self._tracked_layers[layer_name]
            self._null_extras[f'{layer_name}_{method}'] = np.nan

    def __repr__(self):
        msg = "{} starting at {}".format(self.__class__.__name__,
                                         self._start_indices)

        return msg

    def _extract_data_from_cost_h5(self, fh):
        """Extract extra info from cost H5 file. """
        self.transform = rasterio.Affine(*fh.profile['transform'])
        self._full_shape = fh.shape
        self._cost_crs = fh.crs

    @property
    def row_offset(self):
        """
        Offset to apply to row indices to move into clipped array

        Returns
        -------
        int
        """
        offset = self._row_slice.start
        if offset is None:
            offset = 0

        return offset

    @property
    def col_offset(self):
        """
        Offset to apply to column indices to move into clipped array

        Returns
        -------
        int
        """
        offset = self._col_slice.start
        if offset is None:
            offset = 0

        return offset

    @property
    def row(self):
        """
        Row index inside clipped array

        Returns
        -------
        int
        """
        return self._start_indices[0]

    @property
    def col(self):
        """
        Column index inside clipped array

        Returns
        -------
        int
        """
        return self._start_indices[1]

    @property
    def cost(self):
        """
        Tie line costs array

        Returns
        -------
        ndarray
        """
        return self._cost

    @property
    def mcp_cost(self):
        """
        Tie line costs array with barrier costs applied for MCP analysis

        Returns
        -------
        ndarray
        """
        return self._mcp_cost

    @property
    def mcp(self):
        """
        MCP_Geometric instance initialized on mcp_cost array with
        starting point at sc_point

        Returns
        -------
        MCP_Geometric
        """
        if self._mcp is None:
            check = self.mcp_cost[self.row, self.col]
            if check < 0:
                msg = ("Start idx {} does not have a valid cost!"
                       .format((self.row, self.col)))
                raise InvalidMCPStartValueError(msg)

            logger.debug('Building MCP instance for size {}'
                         .format(self.mcp_cost.shape))
            self._mcp = MCP_Geometric(self.mcp_cost)
            self._cumulative_costs, __ = self._mcp.find_costs(
                starts=[(self.row, self.col)])

        return self._mcp

    @property
    def clip_shape(self):
        """
        Shaped of clipped cost raster

        Returns
        -------
        tuple
        """
        if self._clip_shape is None:
            if self._row_slice == slice(None):
                row_shape = self._full_shape[0]
            else:
                row_max = (self._row_slice.stop if self._row_slice.stop
                           else self._full_shape[0])
                row_min = (self._row_slice.start if self._row_slice.start
                           else 0)
                row_shape = row_max - row_min

            if self._col_slice == slice(None):
                col_shape = self._full_shape[1]
            else:
                col_max = (self._col_slice.stop if self._col_slice.stop
                           else self._full_shape[1])
                col_min = (self._col_slice.start if self._col_slice.start
                           else 0)
                col_shape = col_max - col_min

            self._clip_shape = (row_shape, col_shape)

        return self._clip_shape

    @staticmethod
    def _parse_config(xmission_config=None):
        """
        Load Xmission config if needed

        Parameters
        ----------
        config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects, by default None

        Returns
        -------
        XmissionConfig
        """
        if not isinstance(xmission_config, XmissionConfig):
            xmission_config = XmissionConfig(config=xmission_config)

        logger.debug("Xmissing config:\n%s", xmission_config)
        return xmission_config

    def _clip_costs(self, cost_layers, length_invariant_cost_layers=None,
                    barrier_mult=BARRIERS_MULT):
        """Extract clipped cost arrays from exclusion .h5 files

        Parameters
        ----------
        cost_layers : List[str]
            List of layers in H5 that are summed to determine total
            costs raster used for routing. Costs and distances for each
            individual layer are also reported (e.g. wet and dry costs).
            deteremining path using main cost layer.
        length_invariant_cost_layers : List[str] | None, optional
            List of layers in H5 to be added to the cost raster. The
            costs specified by these layers are not scaled with distance
            traversed across the cell (i.e. fixed one-time costs for
            crossing these cells).
        barrier_mult : int, optional
            Multiplier on transmission barrier costs.
            By default, ``100``.
        """
        li_cost_layers = length_invariant_cost_layers or []
        logger.debug("Building cost layer with the folowing inputs:"
                     f"\n\t- cost_layers: {cost_layers}"
                     f"\n\t- length_invariant_cost_layers: {li_cost_layers}"
                     f"\n\t- barrier_mult: {barrier_mult}")
        self._cost = np.zeros(self.clip_shape, dtype=np.float32)
        overlap = np.zeros(self.clip_shape, dtype=np.uint8)
        with ExclusionLayers(self._cost_fpath) as f:
            for cost_layer in cost_layers:
                cost = f[cost_layer, self._row_slice, self._col_slice]
                self._cost += cost
                overlap += cost > 0
                self._cost_layer_map[cost_layer] = cost

            for li_cost_layer in li_cost_layers:
                li_cost = f[li_cost_layer, self._row_slice, self._col_slice]
                overlap += li_cost > 0
                self._li_cost_layer_map[li_cost_layer] = li_cost

            for tracked_layer, method in self._tracked_layers.items():
                if getattr(np, method, None) is None:
                    msg = (f"Did not find method {method!r} in numpy! "
                           f"Skipping tracked layer {tracked_layer!r}")
                    logger.warning(msg)
                    warn(msg)
                    continue
                if tracked_layer not in f.layers:
                    msg = (f"Did not find layer {tracked_layer!r} in cost "
                           f"file {str(self._cost_fpath)!r}. Skipping...")
                    logger.warning(msg)
                    warn(msg)
                    continue

                layer = f[tracked_layer, self._row_slice, self._col_slice]
                self._tracked_layer_map[tracked_layer] = layer

            barrier = f[self._tb_layer_name, self._row_slice, self._col_slice]
            barrier = barrier * barrier_mult

        if (overlap > 1).any():
            all_layers = cost_layers + li_cost_layers
            msg = (f"Found overlap in cost layers: {all_layers}! Some cells "
                   "may contain double-counted costs. If you intentionally "
                   "specified cost layers that overlap, please ignore this "
                   "message. Otherwise, verify that all cost layers are "
                   "mutually exclusive!")
            logger.warning(msg)
            warn(msg)

        self._set_mcp_cost(barrier)

    def _set_mcp_cost(self, barrier):
        """Compute routing costs. """

        self._mcp_cost = self._cost.copy()
        for li_cost_layer in self._li_cost_layer_map.values():
            self._mcp_cost += li_cost_layer

        self._mcp_cost *= 1 + barrier
        self._mcp_cost = np.where(self._mcp_cost <= 0, -1, self._mcp_cost)
        logger.debug("MCP cost min: %.2f, max: %.2f, median: %.2f",
                     np.min(self._mcp_cost), np.max(self._mcp_cost),
                     np.median(self._mcp_cost))

    def _compute_path_length(self, indices: npt.NDArray):
        """
        Compute the total length and cell by cell length of the lease
        cost path defined by 'indices'

        Parameters
        ----------
        indices : ndarray
            n x 2 array of MCP traceback of least cost path

        Returns
        -------
        length : float
            Total length of path in km
        lens : ndarray
            Vector of the distance of the least cost path across each
            cell
        """
        # Use Pythagorean theorem to calculate lengths between cells (km)
        # Use c**2 = a**2 + b**2 to determine length of individual paths
        lens = np.sqrt(np.sum(np.diff(indices, axis=0)**2, axis=1))
        length = np.sum(lens) * self._cell_size / 1000

        # Need to determine distance coming into and out of any cell. Assume
        # paths start and end at the center of a cell. Therefore, distance
        # traveled in the cell is half the distance entering it and half the
        # distance exiting it. Duplicate all lengths, pad 0s on ends for start
        # and end cells, and divide all distance by half.
        lens = np.repeat(lens, 2)
        lens = np.insert(np.append(lens, 0), 0, 0)
        lens = lens / 2

        # Group entrance and exits distance together, and add them
        lens = lens.reshape((int(lens.shape[0] / 2), 2))
        lens = np.sum(lens, axis=1)

        return length, lens

    def least_cost_path(self, end_idx: Tuple[int, int], save_path=False) \
        -> Tuple[float, float, float, float,
                 Optional[Union[Point, LineString]], Dict[str, float]]:
        """
        Find least cost path, its length, and its total (un-barriered)
        cost. Optionally, calculate separate costs for sub-layers, and
        geometry of path.

        Parameters
        ----------
        end_idx : Tuple[int, int]
            (row, col) index of end point to connect and compute least
            cost path to.
        save_path : bool
            Flag to save least cost path as a multi-line geometry.
            By default, ``False``.

        Returns
        -------
        length : float
            Length of path (km).
        cost : float
            Cost of path including terrain and land use multipliers, but
            not barrier costs.
        poi_lat, poi_lon : numpy.float64
            Latitude and longitude of the `end_idx` of the least cost
            path (i.e. the POI/transmission feature that was connected
            to).
        path : shapely.geometry.linestring | None, optional
            Path as a LineString, if `save_path` was set to ``True``.
        cl_results : Dict[str, float]
            Costs and lengths for individual sub-layers.
        """
        row, col = end_idx
        check = (row < 0 or col < 0 or row >= self.clip_shape[0]
                 or col >= self.clip_shape[1])
        if check:
            msg = ('End point ({}, {}) is out side of clipped cost raster '
                   'with shape {}'.format(row, col, self.clip_shape))
            logger.exception(msg)
            raise ValueError(msg)

        check = self.mcp_cost[row, col]
        if check < 0:
            msg = ("End idx {} does not have a valid cost!"
                   .format(end_idx))
            raise LeastCostPathNotFoundError(msg)

        logger.debug("Computing least cost path for end point (%d, %d)",
                     row, col)
        try:
            # Extract path indices
            indices = np.array(self.mcp.traceback((row, col)))
        except ValueError as ex:
            msg = ('Unable to find path from start {} to {}: {}'
                   .format((self.row, self.col), end_idx, ex))
            raise LeastCostPathNotFoundError(msg) from ex

        # Extract costs of cells
        # pylint: disable=unsubscriptable-object
        cell_costs = self.cost[indices[:, 0], indices[:, 1]]
        length, lens = self._compute_path_length(indices)

        # Multiple distance travel through cell by cost and sum it!
        cost = np.sum(cell_costs * lens)

        if self._li_cost_layer_map:
            li_costs = sum(self._li_cost_layer_map.values())
            li_cell_costs = li_costs[indices[:, 0], indices[:, 1]]
            cost += np.sum(li_cell_costs)

        # Determine parial costs for any cost layers
        cl_results = self._compute_by_layer_results(indices, lens, cost)
        cl_results = self._compute_tracked_layer_values(cl_results, indices)

        with ExclusionLayers(self._cost_fpath) as f:
            poi_lat = f['latitude', self._row_slice, self._col_slice][row, col]
            poi_lon = (
                f['longitude', self._row_slice, self._col_slice][row, col])

        if save_path:
            row = indices[:, 0] + self.row_offset
            col = indices[:, 1] + self.col_offset
            x, y = rasterio.transform.xy(self.transform, row, col)
            geom = Point if indices.shape[0] == 1 else LineString
            return (length, cost, poi_lat, poi_lon, geom(list(zip(x, y))),
                    cl_results)

        return (length, cost, poi_lat, poi_lon, None, cl_results)

    def _compute_by_layer_results(self, indices, lens, cost):
        """Compute costs and dists by individual layer. """
        # Determine parial costs for any cost layers
        cl_results: Dict[str, float] = {}
        logger.debug('Calculating partial costs and lengths for: %s',
                     list(self._cost_layer_map.keys()))

        cl_results = _compute_individual_layers_costs_lens(
            self._cost_layer_map, indices, lens, cl_results,
            scale_by_length=True, cell_size=self._cell_size)
        cl_results = _compute_individual_layers_costs_lens(
            self._li_cost_layer_map, indices, lens, cl_results,
            scale_by_length=False, cell_size=self._cell_size)
        test_total_cost = sum(layer
                              for layer_name, layer in cl_results.items()
                              if layer_name.endswith("_cost"))

        # Sanity check that total of cost layers matches full cost raster.
        if round(test_total_cost) != round(cost):
            msg = (f'Sum of cost_layers costs ({test_total_cost:,}) does '
                   f'not equal cost calculated with composite cost layer '
                   f'({cost:,})')
            logger.warning(msg)
            warn(msg)

        return cl_results

    def _compute_tracked_layer_values(self, cl_results, indices):
        """Compute aggregate values over tracked layers. """

        for layer_name, layer_values_arr in self._tracked_layer_map.items():
            layer_values = layer_values_arr[indices[:, 0], indices[:, 1]]
            method = self._tracked_layers[layer_name]
            aggregate = getattr(np, method)(layer_values).astype(float)
            cl_results[f'{layer_name}_{method}'] = aggregate

        return cl_results

    def compute(self, end_indices, save_paths=False):
        """
        Compute least cost paths to given end indices

        Parameters
        ----------
        end_indices : tuple | list
            Tuple (row, col) index or list of (row, col) indices in the
            **clipped** cost array indicating the end location(s) to
            compute least cost path paths to. Paths are computed from
            the `start_indices` that this class was initialized with to
            each of the individual pairs of `end_indices`.
        save_paths : bool, optional
            Flag to save least cost path as a multi-line geometry,
            by default False

        Returns
        -------
        tie_lines : pandas.DataFrame | gpd.GeoDataFrame
            DataFrame of lengths and costs for each path or GeoDataFrame
            of length, cost, and geometry for each path
        """
        if isinstance(end_indices, tuple):
            end_indices = [end_indices]

        lengths = []
        costs = []
        paths = []
        poi_lats = []
        poi_lons = []
        rows = []
        cols = []
        extras = {}
        for end_idx in end_indices:
            rows.append(end_idx[0])
            cols.append(end_idx[1])
            try:
                out = self.least_cost_path(end_idx, save_path=save_paths)
            except LeastCostPathNotFoundError as ex:
                msg = ('Unable to find path from start {} to {}: {}\n'
                       'Skipping...'.format((self.row, self.col), end_idx, ex))
                logger.warning(msg)
                warn(msg)
                out = [None] * 5 + [self._null_extras]

            lengths.append(out[0])
            costs.append(out[1])
            poi_lats.append(out[2])
            poi_lons.append(out[3])
            if save_paths:
                shape = out[4]
                if shape is None:
                    shape = Point(*end_idx)
                paths.append(out[4])

            for k, v in out[-1].items():
                extras.setdefault(k, []).append(v)

        final_output = {'length_km': lengths, 'cost': costs,
                        'poi_lat': poi_lats, 'poi_lon': poi_lons,
                        'row': rows, 'col': cols}
        final_output.update(extras)
        tie_lines = pd.DataFrame(final_output)

        if save_paths:
            tie_lines = gpd.GeoDataFrame(tie_lines, geometry=paths,
                                         crs=self._cost_crs)

        return tie_lines

    @classmethod
    def run(cls, cost_fpath, start_indices, end_indices, cost_layers,
            row_slice, col_slice, xmission_config=None,
            tb_layer_name=BARRIER_H5_LAYER_NAME, barrier_mult=BARRIERS_MULT,
            save_paths=False, length_invariant_cost_layers=None,
            tracked_layers=None, cell_size=CELL_SIZE):
        """
        Compute least cost tie-line path to all features to be connected
        a single supply curve point.

        Parameters
        ----------
        cost_fpath : str
            Full path to HDF5 file containing cost arrays. The cost data
            layers should be named ``"tie_line_costs_{capacity}MW"``,
            where ``capacity`` is an integer value representing the
            capacity of the line (the integer values must matches at
            least one of the integer capacity values represented by the
            keys in the ``power_classes`` portion of the transmission
            config).
        start_indices : tuple
            Tuple of (row_idx, col_idx) in the **clipped** cost array
            (see `row_slice` and `col_slice` definitions below)
            indicating the start position of all paths to compute
            (typically, this is the centroid of the supply curve cell).
            Paths will be computed from this start location to each of
            the ``end_indices``, which are also locations in the cost
            array (typically transmission feature locations).
        cost_layers : List[str]
            List of layers in H5 that are summed to determine total
            costs raster used for routing. Costs and distances for each
            individual layer are also reported (e.g. wet and dry costs).
            deteremining path using main cost layer.
        row_slice, col_slice : slice
            Slices into the cost raster array used to clip the area that
            should be considered when computing a least cost path. This
            can be used to limit the consideration space and speed up
            computation. Note that the start and end indices must
            be given w.r.t. the cost raster that is "clipped" using
            these slices.
        xmission_config : str | dict | XmissionConfig, optional
            Path to transmission config JSON files, dictionary of
            transmission config JSONs, or preloaded XmissionConfig
            objects. If ``None``, the default config is used.
            By default, ``None``.
        tb_layer_name : str, default=:obj:`BARRIER_H5_LAYER_NAME`
            Name of transmission barrier layer in `cost_fpath` file.
            This layer defines the multipliers applied to the cost layer
            to determine LCP routing (but does not actually affect
            output costs). By default, :obj:`BARRIER_H5_LAYER_NAME`.
        barrier_mult : int, optional
            Multiplier applied to the cost data to create transmission
            barrier costs. By default, ``100``.
        save_paths : bool, optional
            Flag to save least cost path as a multi-line geometry.
            By default, ``False``.
        length_invariant_cost_layers : List[str] | None, optional
            List of layers in H5 to be added to the cost raster. The
            costs specified by these layers are not scaled with distance
            traversed across the cell (i.e. fixed one-time costs for
            crossing these cells).
        tracked_layers : dict, optional
            Dictionary mapping layer names to strings, where the strings
            are numpy methods that should be applied to the layer along
            the LCP. For example,
            ``tracked_layers={'layer_1': 'mean', 'layer_2': 'max}``
            would report the average of ``layer_1`` values along the
            least cost path and the max of ``layer_2`` values along the
            least cost path. Examples of numpy methods (non-exhaustive):

                - mean
                - max
                - min
                - mode
                - median
                - std

            By default, ``None``, which does not track any extra layers.
        cell_size : int, optional
            Side length of each cell, in meters. Cells are assumed to be
            square. By default, :obj:`CELL_SIZE`.

        Returns
        -------
        tie_lines : pandas.DataFrame | gpd.GeoDataFrame
            DataFrame of lengths and costs for each path or GeoDataFrame
            of length, cost, and geometry for each path
        """
        ts = time.time()
        tlc = cls(cost_fpath, start_indices, cost_layers, row_slice,
                  col_slice, xmission_config=xmission_config,
                  tb_layer_name=tb_layer_name, barrier_mult=barrier_mult,
                  length_invariant_cost_layers=length_invariant_cost_layers,
                  tracked_layers=tracked_layers, cell_size=cell_size)

        tie_lines = tlc.compute(end_indices, save_paths=save_paths)

        logger.debug('Least Cost tie-line computed in {:.4f} min'
                     .format((time.time() - ts) / 60))

        return tie_lines


class TransCapCosts(TieLineCosts):
    """
    Compute total transmission capital cost
    (least-cost tie-line cost + connection cost) for all features to be
    connected a single supply curve point
    """
    _CHECK_FOR_INVALID_REGION = True

    def __init__(self, cost_fpath, sc_point, features, capacity_class,
                 cost_layers, radius=None, xmission_config=None,
                 tb_layer_name=BARRIER_H5_LAYER_NAME,
                 barrier_mult=BARRIERS_MULT,
                 iso_regions_layer_name=ISO_H5_LAYER_NAME,
                 length_invariant_cost_layers=None, tracked_layers=None,
                 cell_size=CELL_SIZE):
        """
        Parameters
        ----------
        cost_fpath : str
            Full path of .h5 file with cost arrays
        sc_point : gpd.GeoSeries
            Supply Curve Point meta data
        features : pandas.DataFrame
            Table of transmission features to connect to supply curve
            point. Must have "row" and "col" columns that point to the
            indexs of the feature **in the original cost array**. Must
            also have a "category" column that distinguishes between
            substations and transmission lines.
        capacity_class : int | str
            Transmission feature ``capacity_class`` class. Used to look
            up connection costs.
        cost_layers : List[str]
            List of layers in H5 that are summed to determine total
            costs raster used for routing. Costs and distances for each
            individual layer are also reported (e.g. wet and dry costs).
            deteremining path using main cost layer.
        radius : int, optional
            Radius around sc_point to clip cost to, by default None
        xmission_config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects, by default None
        tb_layer_name : str, default=:obj:`BARRIER_H5_LAYER_NAME`
            Name of transmission barrier layer in `cost_fpath` file.
            This layer defines the multipliers applied to the cost layer
            to determine LCP routing (but does not actually affect
            output costs). By default, :obj:`BARRIER_H5_LAYER_NAME`.
        barrier_mult : int, optional
            Multiplier on transmission barrier costs, by default 100
        iso_regions_layer_name : str, default=:obj:`ISO_H5_LAYER_NAME`
            Name of ISO regions layer in `cost_fpath` file. The layer
            maps pixels to ISO region ID's (1, 2, 3, 4, etc.) .
            By default, :obj:`ISO_H5_LAYER_NAME`.
        length_invariant_cost_layers : List[str] | None, optional
            List of layers in H5 to be added to the cost raster. The
            costs specified by these layers are not scaled with distance
            traversed across the cell (i.e. fixed one-time costs for
            crossing these cells).
        tracked_layers : dict, optional
            Dictionary mapping layer names to strings, where the strings
            are numpy methods that should be applied to the layer along
            the LCP. For example,
            ``tracked_layers={'layer_1': 'mean', 'layer_2': 'max}``
            would report the average of ``layer_1`` values along the
            least cost path and the max of ``layer_2`` values along the
            least cost path. Examples of numpy methods (non-exhaustive):

                - mean
                - max
                - min
                - mode
                - median
                - std

            By default, ``None``, which does not track any extra layers.
        cell_size : int, optional
            Side length of each cell, in meters. Cells are assumed to be
            square. By default, :obj:`CELL_SIZE`.
        """
        self._sc_point = sc_point
        self._region_layer = None
        self._iso_regions_layer_name = iso_regions_layer_name
        start_indices, row_slice, col_slice = self._get_clipping_slices(
            cost_fpath, sc_point[['row', 'col']].values, radius=radius)
        licl = length_invariant_cost_layers
        super().__init__(cost_fpath, start_indices, cost_layers, row_slice,
                         col_slice, xmission_config=xmission_config,
                         tb_layer_name=tb_layer_name,
                         barrier_mult=barrier_mult,
                         length_invariant_cost_layers=licl,
                         tracked_layers=tracked_layers,
                         cell_size=cell_size)
        self._capacity_class = self._config._parse_cap_class(capacity_class)
        self._features = self._prep_features(features)
        self._clip_mask = None

    def _extract_data_from_cost_h5(self, fh):
        """Extract extra info from cost H5 file. """
        super()._extract_data_from_cost_h5(fh)
        self._region_layer = fh[self._iso_regions_layer_name,
                                self._row_slice, self._col_slice]

    @property
    def sc_point(self):
        """
        Supply curve point data:
        - gid
        - lat
        - lon
        - idx (row, col)

        Returns
        -------
        pandas.Series
        """
        return self._sc_point

    @property
    def sc_point_gid(self):
        """
        Supply curve point gid

        Returns
        -------
        int
        """
        return int(self.sc_point['sc_point_gid'])

    @property
    def features(self) -> pd.DataFrame:
        """
        Table of transmission features

        Returns
        -------
        pandas.DataFrame
        """
        return self._features

    @property
    def clip_mask(self):
        """
        Polygon used to clip transmission lines to the clipped raster
        bounds

        Returns
        -------
        shapely.Polygon
        """
        if self._clip_mask is None:
            # pylint: disable=using-constant-test
            row_bounds = [self._row_slice.start
                          if self._row_slice.start else 0,
                          self._row_slice.stop - 1
                          if self._row_slice.stop else self.clip_shape[0] - 1]
            col_bounds = [self._col_slice.start
                          if self._col_slice.start else 0,
                          self._col_slice.stop - 1
                          if self._col_slice.stop else self.clip_shape[1] - 1]
            x, y = rasterio.transform.xy(self.transform, row_bounds,
                                         col_bounds)
            self._clip_mask = Polygon([[x[0], y[0]],
                                       [x[1], y[0]],
                                       [x[1], y[1]],
                                       [x[0], y[1]],
                                       [x[0], y[0]]])

        return self._clip_mask

    @property
    def capacity_class(self):
        """
        SC point capacity class

        Returns
        -------
        str
        """
        return self._capacity_class

    @property
    def tie_line_voltage(self):
        """
        Tie line voltage in kV

        Returns
        -------
        int
        """
        return self._config.capacity_to_kv(self.capacity_class)

    @staticmethod
    def _get_clipping_slices(cost_fpath, sc_point_idx, radius=None):
        """
        Get array slices for clipped area around SC point (row, col)
        index

        Parameters
        ----------
        cost_fpath : str
            Full path of .h5 file with cost arrays
        row : int
            SC point row index
        col : int
            SC point column index
        radius : int, optional
            Radius around sc_point to clip cost to, by default None

        Returns
        -------
        start_indices : tuple
            Start index in clipped raster space
        row_slice : slice
            Row start, stop indices for clipped cost array
        col_slice : slice
            Column start, stop indices for clipped cost array
        """
        with ExclusionLayers(cost_fpath) as f:
            shape = f.shape

        if radius is not None:
            row, col = sc_point_idx
            row_min = max(row - radius, 0)
            row_max = min(row + radius, shape[0])
            col_min = max(col - radius, 0)
            col_max = min(col + radius, shape[1])

            start_indices = (row - row_min, col - col_min)
        else:
            start_indices = sc_point_idx
            row_min, row_max = None, None
            col_min, col_max = None, None

        row_slice = slice(row_min, row_max)
        col_slice = slice(col_min, col_max)

        return start_indices, row_slice, col_slice

    def _calc_xformer_cost(self, features):
        """
        Compute transformer costs in $/MW for needed features, all
        others will be 0

        Parameters
        ----------
        features : pd.DataFrame
            Table of transmission features to compute transformer costs
            for
        tie_line_voltage : int
            Tie-line voltage in kV
        config : str | dict | XmissionConfig
            Transmission configuration

        Returns
        -------
        xformer_costs : ndarray
            vector of $/MW transformer costs
        """

        mask = features['category'] == SUBSTATION_CAT
        mask &= features['max_volts'] < self.tie_line_voltage
        if np.any(mask):
            msg = ('Voltages for substations {} do not exceed tie-line '
                   'voltage of {}'
                   .format(features.loc[mask, 'trans_gid'].values,
                           self.tie_line_voltage))
            logger.error(msg)
            raise RuntimeError(msg)

        xformer_costs = np.zeros(len(features))
        for volts, df in features.groupby('min_volts'):
            idx = df.index.values
            xformer_costs[idx] = self._config.xformer_cost(
                volts, self.tie_line_voltage)

        mask = features['category'] == TRANS_LINE_CAT
        mask &= features['max_volts'] < self.tie_line_voltage
        xformer_costs[mask] = 0

        mask = features['min_volts'] <= self.tie_line_voltage
        xformer_costs[mask] = 0

        mask = features['region'] == self._config['iso_lookup']['TEPPC']
        xformer_costs[mask] = 0

        return xformer_costs

    def _calc_sub_upgrade_cost(self, features):
        """
        Compute substation upgrade costs for needed features, all others
        will be 0

        Parameters
        ----------
        features : pd.DataFrame
            Table of transmission features to compute transformer costs
            for

        Returns
        -------
        sub_upgrade_costs : ndarray
            Substation upgrade costs in $
        """

        sub_upgrade_cost = np.zeros(len(features))
        if self._CHECK_FOR_INVALID_REGION and np.any(features['region'] == 0):
            mask = features['region'] == 0
            msg = ('Features {} have an invalid region! Region must != 0!'
                   .format(features.loc[mask, 'trans_gid'].values))
            logger.error(msg)
            raise RuntimeError(msg)

        mask = features['category'].isin([SUBSTATION_CAT, LOAD_CENTER_CAT])
        if np.any(mask):
            for region, df in features.loc[mask].groupby('region'):
                idx = df.index.values
                sub_upgrade_cost[idx] = self._config.sub_upgrade_cost(
                    region, self.tie_line_voltage)

        return sub_upgrade_cost

    def _calc_new_sub_cost(self, features):
        """
        Compute new substation costs for needed features, all others
        will be 0

        Parameters
        ----------
        features : pd.DataFrame
            Table of transmission features to compute transformer costs
            for

        Returns
        -------
        new_sub_cost : ndarray
            new substation costs in $
        """

        new_sub_cost = np.zeros(len(features))
        if self._CHECK_FOR_INVALID_REGION and np.any(features['region'] == 0):
            mask = features['region'] == 0
            msg = ('Features {} have an invalid region! Region must != 0!'
                   .format(features.loc[mask, 'trans_gid'].values))
            logger.error(msg)
            raise RuntimeError(msg)

        mask = features['category'] == TRANS_LINE_CAT
        if np.any(mask):
            for region, df in features.loc[mask].groupby('region'):
                idx = df.index.values
                new_sub_cost[idx] = self._config.new_sub_cost(
                    region, self.tie_line_voltage)

        return new_sub_cost

    def _prep_features(self, features):
        """
        Shift feature row and col indices of transmission features from
        the global domain to the clipped raster, clip transmission lines
        to clipped array and find nearest point

        Parameters
        ----------
        features : pandas.DataFrame
            Table of transmission features

        Returns
        -------
        features : pandas.DataFrame
            Transmission features with row/col indices shifted to
            clipped raster
        """
        mapping = {'gid': 'trans_gid', 'trans_gids': 'trans_line_gids'}
        features = features.rename(columns=mapping).drop(columns='dist',
                                                         errors='ignore')

        if self.row_offset is not None:
            features['row'] -= self.row_offset

        if self.col_offset is not None:
            features['col'] -= self.col_offset

        return features.reset_index(drop=True)

    def _get_trans_line_idx(self, trans_line, clip=False):
        """
        Map the nearest point on each transmission lines to the cost
        raster

        Parameters
        ----------
        trans_lines : GeoPandas.GeoSeries
            Transmission lines to be connected to each supply curve
            point, the nearest point on each line needs to be mapped to
            the cost raster grid in order to compute the least cost path
        clip : bool
            Flag to clip the transmission lines to the cost raster
            domain

        Returns
        -------
        [row, col] : list
            Row, col index of the nearest point on the transmission line
            to the supply curve point, used for least cost path
        """
        if clip:
            logger.debug("Clipping transmission line {} to raster domain"
                         .format(trans_line['trans_gid']))
            clipped_trans_line = {'geometry': trans_line['geometry']}
            clipped_trans_line = gpd.clip(gpd.GeoSeries(clipped_trans_line),
                                          self.clip_mask)
            if len(clipped_trans_line) == len(trans_line):
                trans_line = clipped_trans_line

        point, _ = nearest_points(trans_line['geometry'],
                                  self.sc_point['geometry'])
        row, col = rasterio.transform.rowcol(self.transform, point.x, point.y)

        row -= self.row_offset
        col -= self.col_offset
        if not clip:
            clip = (row < 0 or row >= self.clip_shape[0]
                    or col < 0 or col >= self.clip_shape[1])
            if clip:
                row, col = self._get_trans_line_idx(trans_line, clip=clip)
        else:
            row = min(max(row, 0), self.clip_shape[0] - 1)
            col = min(max(col, 0), self.clip_shape[1] - 1)

        return [row, col]

    def _check_tline_voltage(self, cost, feat):
        """Return large cost if tie line voltage is too low. """
        if feat['max_volts'] < self.tie_line_voltage:
            msg = ('Tie-line {} voltage of {}kV is less than tie line '
                   'voltage of {}kV.'
                   .format(feat['trans_gid'], feat['max_volts'],
                           self.tie_line_voltage))
            logger.debug(msg)
            cost = 1e12

        return cost

    def compute_tie_line_costs(self,  # noqa: C901
                               min_line_length=MINIMUM_SPUR_DIST_KM,
                               save_paths=False,
                               ) -> Union[pd.DataFrame, gpd.GeoDataFrame]:
        """
        Compute least cost path and distance between supply curve point
        and every transmission feature

        Parameters
        ----------
        min_line_length : float, optional
            Minimum line length in km, by default 0
        save_paths : bool, optional
            Flag to save least cost path as a multi-line geometry,
            by default False

        Returns
        -------
        tie_line_costs : gpd.GeoDataFrame
            Updated table of transmission features with the tie-line
            cost and distance added
        """
        features = self.features.copy()
        features['raw_line_cost'] = None
        features['dist_km'] = None
        features['poi_gid'] = None
        if save_paths:
            paths = []

        for col, val in self._null_extras.items():
            features[col] = val

        logger.info('Computing paths to %d features for SC Point %d',
                    len(features), self.sc_point_gid)
        for iter_ind, (index, feat) in enumerate(features.iterrows(), start=1):
            logger.debug('Determining path length and cost to feature:\n%s',
                         feat)

            t_line = feat['category'] == TRANS_LINE_CAT
            if t_line:
                feat_idx = self._get_trans_line_idx(feat)
            else:
                feat_idx = feat[['row', 'col']].values

            logger.debug('Feat index is: %s', feat_idx)

            row, col = feat_idx
            region = self._region_layer[row, col]
            row += self.row_offset
            col += self.col_offset
            poi_gid = self._full_shape[1] * row + col
            logger.debug('Adding row, col, poi_gid, region: %d, %d, %d, %s',
                         row, col, poi_gid, str(region))

            features.loc[index, 'row'] = row
            features.loc[index, 'col'] = col
            features.loc[index, 'poi_gid'] = poi_gid
            features.loc[index, 'region'] = region

            try:
                result = self.least_cost_path(feat_idx, save_path=save_paths)
                (length, cost, poi_lat, poi_lon, path, cl_results) = result

                if t_line:
                    cost = self._check_tline_voltage(cost, feat)

                if length < min_line_length:
                    msg = ('Tie-line length {} will be increased to the '
                           'minimum allowed line length: {}.'
                           .format(length, min_line_length))
                    logger.debug(msg)

                    min_mult = (1 if np.isclose(length, 0)
                                else min_line_length / length)
                    cost = cost * min_mult
                    length = min_line_length

                features.loc[index, 'dist_km'] = length
                features.loc[index, 'raw_line_cost'] = cost
                features.loc[index, 'poi_lat'] = poi_lat
                features.loc[index, 'poi_lon'] = poi_lon
                if save_paths:
                    paths.append(path)

                # Append any cost layer values
                for key, value in cl_results.items():
                    logger.debug("Adding %s to output", key)
                    features.loc[index, key] = value

            except LeastCostPathNotFoundError as ex:
                msg = ("Could not connect SC point {} to transmission feature "
                       "{}: {}"
                       .format(self.sc_point_gid, feat['trans_gid'], ex))
                logger.debug(msg)
                if t_line:
                    features.loc[index, 'raw_line_cost'] = 1e12
                if save_paths:
                    paths.append(self._sc_point.geometry)
            except InvalidMCPStartValueError:
                raise
            except Exception:
                logger.exception('Could not connect SC point %d to '
                                 'transmission features!', self.sc_point_gid)
                raise

            logger.info('Processed %d out of %d features (%.2f%%) for SC '
                        'point %d',
                        iter_ind, len(features),
                        iter_ind / len(features) * 100,
                        self.sc_point_gid)

        if save_paths:
            features = gpd.GeoDataFrame(features, geometry=paths,
                                        crs=self._cost_crs)

        for int_col in ["row", "col", "poi_gid"]:
            features[int_col] = features[int_col].astype("int64")
        for float_col in ["raw_line_cost", "dist_km"]:
            features[float_col] = features[float_col].astype("float64")
        return features

    def compute_connection_costs(self, features=None,
                                 length_mult_kind="linear"):
        """
        Calculate connection costs for tie lines

        Parameters
        ----------
        features : str, optional
            Optional features input. If ``None``, features held by this
            object are used.
        length_mult_kind : {"step", "linear"}, default="linear"
            Type of length multiplier calcualtion. "step" computes
            length multipliers using a step function, while "linear"
            computes the length multiplier using a linear interpolation
            between 0 amd 10 mile spur-line lengths.
            By default, ``"linear"``.

        Returns
        -------
        features : pd.DataFrame
            Updated table of transmission features with the connection
            costs added
        """
        if features is None:
            features = self.features.copy()

        features = features.reset_index(drop=True)
        features = _compute_length_mult(features, kind=length_mult_kind)

        features['tie_line_cost'] = (features['raw_line_cost']
                                     * features['length_mult'])

        # Transformer costs
        features['xformer_cost_per_mw'] = self._calc_xformer_cost(features)
        capacity = int(self.capacity_class.strip('MW'))
        features['xformer_cost'] = (features['xformer_cost_per_mw']
                                    * capacity)

        # Substation costs
        features['sub_upgrade_cost'] = self._calc_sub_upgrade_cost(features)
        features['new_sub_cost'] = self._calc_new_sub_cost(features)

        # Sink costs
        mask = features['category'] == SINK_CAT
        features.loc[mask, 'new_sub_cost'] = SINK_CONNECTION_COST

        # Total cost
        features['connection_cost'] = (features['xformer_cost']
                                       + features['sub_upgrade_cost']
                                       + features['new_sub_cost'])

        return features

    def compute(self, min_line_length=MINIMUM_SPUR_DIST_KM, save_paths=False,
                simplify_geo=None, length_mult_kind="linear"):
        """
        Compute Transmission capital cost of connecting SC point to
        transmission features.
        trans_cap_cost = tie_line_cost + connection_cost

        Parameters
        ----------
        min_line_length : float, optional
            Minimum line length in km, by default 0
        save_paths : bool, optional
            Flag to save least cost path as a multi-line geometry,
            by default False
        simplify_geo : float | None, optional
            If float, simplify geometries using this value
        length_mult_kind : {"step", "linear"}, default="linear"
            Type of length multiplier calcualtion. "step" computes
            length multipliers using a step function, while "linear"
            computes the length multiplier using a linear interpolation
            between 0 amd 10 mile spur-line lengths.
            By default, ``"linear"``.

        Returns
        -------
        features : pd.DataFrame | gpd.GeoDataFrame
            Transmission table with tie-line costs and distances and
            connection costs added. Includes paths if
            ``save_paths == True``
        """
        features = self.compute_tie_line_costs(min_line_length=min_line_length,
                                               save_paths=save_paths)

        mask = features['raw_line_cost'].isna()
        if mask.any():
            msg = ("The following features could not be connected to SC point "
                   "{}:\n{}".format(self.sc_point_gid,
                                    features.loc[mask, 'trans_gid']))
            logger.warning(msg)
            warn(msg)
            features = features.loc[~mask]

        features = self.compute_connection_costs(
            features=features, length_mult_kind=length_mult_kind)

        features['trans_cap_cost'] = (features['tie_line_cost']
                                      + features['connection_cost'])
        if not save_paths:
            features = features.drop(columns=['geometry'],
                                     errors='ignore').reset_index(drop=True)

        features['sc_row_ind'] = self.sc_point['sc_row_ind']
        features['sc_col_ind'] = self.sc_point['sc_col_ind']
        features['sc_point_gid'] = self.sc_point_gid

        if save_paths and simplify_geo:
            features.geometry = features.geometry.simplify(simplify_geo)

        return features

    @classmethod
    def run(cls, cost_fpath, sc_point, features, capacity_class, cost_layers,
            radius=None, xmission_config=None,
            tb_layer_name=BARRIER_H5_LAYER_NAME, barrier_mult=BARRIERS_MULT,
            iso_regions_layer_name=ISO_H5_LAYER_NAME,
            min_line_length=MINIMUM_SPUR_DIST_KM, save_paths=False,
            simplify_geo=None, length_invariant_cost_layers=None,
            tracked_layers=None, length_mult_kind="linear",
            cell_size=CELL_SIZE):
        """
        Compute Transmission capital cost of connecting SC point to
        transmission features.
        trans_cap_cost = tie_line_cost + connection_cost

        Parameters
        ----------
        cost_fpath : str
            Full path of .h5 file with cost arrays
        sc_point : gpd.GeoSeries
            Supply Curve Point meta data
        features : pandas.DataFrame
            Table of transmission features to connect to supply curve
            point. Must have "row" and "col" columns that point to the
            indexs of the feature **in the original cost array**. Must
            also have a "category" column that distinguishes between
            substations and transmission lines.
        capacity_class : int | str
            Transmission feature capacity_class class
        cost_layers : List[str]
            List of layers in H5 that are summed to determine total
            costs raster used for routing. Costs and distances for each
            individual layer are also reported (e.g. wet and dry costs).
            deteremining path using main cost layer.
        radius : int, optional
            Radius around sc_point to clip cost to, by default None
        xmission_config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects, by default None
        tb_layer_name : str, default=:obj:`BARRIER_H5_LAYER_NAME`
            Name of transmission barrier layer in `cost_fpath` file.
            This layer defines the multipliers applied to the cost layer
            to determine LCP routing (but does not actually affect
            output costs). By default, :obj:`BARRIER_H5_LAYER_NAME`.
        barrier_mult : int, optional
            Multiplier on transmission barrier costs, by default 100
        iso_regions_layer_name : str, default=:obj:`ISO_H5_LAYER_NAME`
            Name of ISO regions layer in `cost_fpath` file. The layer
            maps pixels to ISO region ID's (1, 2, 3, 4, etc.) .
            By default, :obj:`ISO_H5_LAYER_NAME`.
        min_line_length : float, optional
            Minimum line length in km, by default 0
        save_paths : bool, optional
            Flag to save least cost path as a multi-line geometry,
            by default False
        simplify_geo : float | None, optional
            If float, simplify geometries using this value
        length_invariant_cost_layers : List[str] | None, optional
            List of layers in H5 to be added to the cost raster. The
            costs specified by these layers are not scaled with distance
            traversed across the cell (i.e. fixed one-time costs for
            crossing these cells).
        tracked_layers : dict, optional
            Dictionary mapping layer names to strings, where the strings
            are numpy methods that should be applied to the layer along
            the LCP. For example,
            ``tracked_layers={'layer_1': 'mean', 'layer_2': 'max}``
            would report the average of ``layer_1`` values along the
            least cost path and the max of ``layer_2`` values along the
            least cost path. Examples of numpy methods (non-exhaustive):

                - mean
                - max
                - min
                - mode
                - median
                - std

            By default, ``None``, which does not track any extra layers.
        length_mult_kind : {"step", "linear"}, default="linear"
            Type of length multiplier calcualtion. "step" computes
            length multipliers using a step function, while "linear"
            computes the length multiplier using a linear interpolation
            between 0 amd 10 mile spur-line lengths.
            By default, ``"linear"``.
        cell_size : int, optional
            Side length of each cell, in meters. Cells are assumed to be
            square. By default, :obj:`CELL_SIZE`.

        Returns
        -------
        features : pd.DataFrame | gpd.GeoDataFrame | None
            Transmission table with tie-line costs and distances and
            connection costs added. Will include paths if
            ``save_paths == True``
        """
        ts = time.time()
        logger.debug('Processing sc_point {}, {}, save_paths={}'
                     .format(sc_point.sc_point_gid, sc_point.geometry,
                             save_paths))

        try:
            licl = length_invariant_cost_layers
            tcc = cls(cost_fpath, sc_point, features, capacity_class,
                      cost_layers, radius=radius,
                      xmission_config=xmission_config,
                      tb_layer_name=tb_layer_name,
                      barrier_mult=barrier_mult,
                      iso_regions_layer_name=iso_regions_layer_name,
                      tracked_layers=tracked_layers,
                      length_invariant_cost_layers=licl,
                      cell_size=cell_size)

            features = tcc.compute(min_line_length=min_line_length,
                                   save_paths=save_paths,
                                   simplify_geo=simplify_geo,
                                   length_mult_kind=length_mult_kind)
            logger.debug('Least Cost transmission costs computed for '
                         'SC point {} in {:.4f}s'
                         .format(tcc.sc_point_gid, time.time() - ts))
        except InvalidMCPStartValueError as ex:
            features = None
            msg = ('Could not connect SC point {} to transmission features: {}'
                   .format(sc_point['sc_point_gid'], ex))
            logger.debug(msg)
        except Exception as ex:
            msg = ('Failed to connect SC point {} to transmission features: {}'
                   .format(sc_point['sc_point_gid'], ex))
            logger.exception(msg)
            raise

        return features


class RegionalTransCapCosts(TransCapCosts):
    """Compute tie-line costs when connections are limitred to a region.

    This class also allows for costs for region values of 0.
    Additionally, the `trans_gid`, `min_volt` and `max_volt` columns are
    no longer required in the features input.
    """
    TRANSFORMER_COST_VOLTAGE = 69
    _CHECK_FOR_INVALID_REGION = False

    def _get_trans_line_idx(self, trans_line):
        """Get the cheapest point on each transmission line.

        Parameters
        ----------
        trans_lines : GeoPandas.GeoSeries
            Transmission lines to be connected to each supply curve
            point.

        Returns
        -------
        [row, col] : list
            Row, col index of the cheapest point on the transmission
            line to route to from the supply curve point.
        """
        window = rasterio.windows.from_bounds(*self.clip_mask.bounds,
                                              transform=self.transform)
        window_transform = rasterio.windows.transform(window=window,
                                                      transform=self.transform)

        mask = rasterio.features.geometry_mask([trans_line.geometry],
                                               out_shape=self.cost.shape,
                                               transform=window_transform,
                                               invert=True)
        rows, cols = np.where(mask)

        __ = self.mcp  # ensures self._cumulative_costs is not `None`
        point_ind = np.argmin(self._cumulative_costs[rows, cols])

        return [rows[point_ind], cols[point_ind]]

    def _check_tline_voltage(self, cost, *__, **___):
        """No adjustments. """
        return cost

    def _calc_xformer_cost(self, features):
        """Compute transformer costs in $/MW

        Parameters
        ----------
        features : pd.DataFrame
            Table of transmission features to compute transformer costs
            for

        Returns
        -------
        xformer_costs : ndarray
            vector of $/MW transformer costs
        """
        cost = self._config.xformer_cost(self.TRANSFORMER_COST_VOLTAGE,
                                         self.tie_line_voltage)
        xformer_costs = np.full(len(features), cost)

        mask = features['region'] == self._config['iso_lookup']['TEPPC']
        xformer_costs[mask] = 0

        return xformer_costs


class ReinforcementLineCosts(TieLineCosts):
    """
    Compute Least Cost Reinforcement Line cost from substations to
    network nodes.

    The reinforcement line path will attempt to follow existing
    transmission lines for as long as possible. Costs are calculated
    using half of the greenfield cost of the transmission line that is
    being traced. If the reinforcement path travels along two different
    line voltages, corresponding costs are used for each portion of the
    path. In the case that the path must cross a region with no existing
    transmission lines to reach the destination, half (50%) of the
    greenfield cost of the ``cost_layers`` input is used.
    """
    def __init__(self, transmission_lines, cost_fpath, start_indices,
                 capacity_class, cost_layers, row_slice, col_slice,
                 xmission_config=None, tb_layer_name=BARRIER_H5_LAYER_NAME,
                 barrier_mult=BARRIERS_MULT,
                 length_invariant_cost_layers=None, tracked_layers=None,
                 cell_size=CELL_SIZE):
        """

        Parameters
        ----------
        transmission_lines : dict
            Dictionary where the keys are the names of cost layers in
            the cost HDF5 file and values are **clipped** arrays with
            the corresponding existing transmission lines rastered into
            them (i.e. array value is 1 at a pixel if there is a
            transmission line, otherwise 0). These arrays will be used
            to compute the reinforcement costs along existing
            transmission lines of differing voltages.
        cost_fpath : str
            Full path of .h5 file with cost arrays.
        start_indices : tuple
            Tuple of (row_idx, col_idx) in the **clipped** cost array
            indicating the start position of all reinforcement line
            paths to compute (typically, this is the location of the
            network node in the BA). Paths will be computed from this
            start location to each of the `end_indices`, which are also
            locations in the cost array (typically substations within
            the BA of the network node).
        capacity_class : int | str
            Capacity class of the 'base' greenfield costs layer. Costs
            will be scaled by the capacity corresponding to this class
            to report reinforcement costs as $/MW.
        cost_layers : List[str]
            List of layers in H5 that are summed to determine total
            'base' greenfield costs raster used for routing. 'Base'
            greenfield costs are only used if the reinforcement path
            *must* deviate from existing transmission lines. Typically,
            a capacity class of 400 MW (230kV transmission line) is used
            for the base greenfield costs.
        row_slice, col_slice : slice
            Row and column slices into the cost array representing the
            window to compute reinforcement line path within.
        xmission_config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects.
            By default, ``None``.
        tb_layer_name : str, default=:obj:`BARRIER_H5_LAYER_NAME`
            Name of transmission barrier layer in `cost_fpath` file.
            This layer defines the multipliers applied to the cost layer
            to determine LCP routing (but does not actually affect
            output costs). By default, :obj:`BARRIER_H5_LAYER_NAME`.
        barrier_mult : int, optional
            Multiplier on transmission barrier costs.
            By default, ``100``.
        length_invariant_cost_layers : List[str] | None, optional
            List of layers in H5 to be added to the 'base' greenfield
            cost raster. The costs specified by these layers are not
            scaled with distance traversed across the cell (i.e. fixed
            one-time costs for crossing these cells).
        tracked_layers : dict, optional
            Dictionary mapping layer names to strings, where the strings
            are numpy methods that should be applied to the layer along
            the LCP. For example,
            ``tracked_layers={'layer_1': 'mean', 'layer_2': 'max}``
            would report the average of ``layer_1`` values along the
            least cost path and the max of ``layer_2`` values along the
            least cost path. Examples of numpy methods (non-exhaustive):

                - mean
                - max
                - min
                - mode
                - median
                - std

            By default, ``None``, which does not track any extra layers.
        cell_size : int, optional
            Side length of each cell, in meters. Cells are assumed to be
            square. By default, :obj:`CELL_SIZE`.
        """
        licl = length_invariant_cost_layers
        super().__init__(cost_fpath=cost_fpath, start_indices=start_indices,
                         cost_layers=cost_layers,
                         row_slice=row_slice, col_slice=col_slice,
                         xmission_config=xmission_config,
                         tb_layer_name=tb_layer_name,
                         barrier_mult=barrier_mult,
                         length_invariant_cost_layers=licl,
                         tracked_layers=tracked_layers,
                         cell_size=cell_size)
        self._null_extras = {}

        capacity_class = self._config._parse_cap_class(capacity_class)
        line_cap_mw = self._config['power_classes'][capacity_class]
        self._cost = self._cost / line_cap_mw
        with ExclusionLayers(cost_fpath) as f:
            for capacity_mw, lines in transmission_lines.items():
                t_lines = np.where(lines)
                cost_layer = 'tie_line_costs_{}MW'.format(capacity_mw)
                costs = f[cost_layer, row_slice, col_slice][t_lines]
                # allow crossing of barriers along existing transmission lines
                costs[costs <= 0] = 1
                self._mcp_cost[t_lines] = costs * 1e-9  # 0 not allowed
                self._cost[t_lines] = costs / capacity_mw

    def _compute_by_layer_results(self, *__, **___):
        """By-layer results not supported for reinforcement run. """
        return {}

    def _compute_tracked_layer_values(self, cl_results, *__, **___):
        """Tracked layer results not supported for reinforcement run.  """
        return cl_results

    @classmethod
    def run(cls, transmission_lines, cost_fpath, start_indices, end_indices,
            capacity_class, cost_layers, row_slice, col_slice,
            xmission_config=None, tb_layer_name=BARRIER_H5_LAYER_NAME,
            barrier_mult=BARRIERS_MULT, save_paths=False,
            length_invariant_cost_layers=None, tracked_layers=None,
            cell_size=CELL_SIZE):
        """
        Compute reinforcement line path to all features to be connected
        a single supply curve point.

        Parameters
        ----------
        transmission_lines : dict
            Dictionary where the keys are the names of cost layers in
            the cost HDF5 file and values are **clipped** arrays with
            the corresponding existing transmission lines rastered into
            them (i.e. array value is 1 at a pixel if there is a
            transmission line, otherwise 0). These arrays will be used
            to compute the reinforcement costs along existing
            transmission lines of differing voltages.
        cost_fpath : str
            Full path of .h5 file with cost arrays.
        start_indices : tuple
            Tuple of (row_idx, col_idx) in the **clipped** cost array
            indicating the start position of all reinforcement line
            paths to compute (typically, this is the location of the
            network node in the BA). Paths will be computed from this
            start location to each of the `end_indices`, which are also
            locations in the cost array (typically substations within
            the BA of the network node).
        end_indices : tuple | list
            Tuple (row, col) index or list of (row, col) indices in the
            **clipped** cost array indicating the end location(s) to
            compute reinforcement line paths to (typically substations
            within a single BA). Paths are computed from the
            `start_indices` (typically the network node of the BA) to
            each of the individual pairs of `end_indices`.
        capacity_class : int | str
            Capacity class of the 'base' greenfield costs layer. Costs
            will be scaled by the capacity corresponding to this class
            to report reinforcement costs as $/MW.
        cost_layers : List[str]
            List of layers in H5 that are summed to determine total
            'base' greenfield costs raster used for routing. 'Base'
            greenfield costs are only used if the reinforcement path
            *must* deviate from existing transmission lines. Typically,
            a capacity class of 400 MW (230kV transmission line) is used
            for the base greenfield costs.
        row_slice, col_slice : slice
            Row and column slices into the cost array representing the
            window to compute reinforcement line path within.
        xmission_config : str | dict | XmissionConfig, optional
            Path to Xmission config .json, dictionary of Xmission config
            .jsons, or preloaded XmissionConfig objects.
            By default, ``None``.
        tb_layer_name : str, default=:obj:`BARRIER_H5_LAYER_NAME`
            Name of transmission barrier layer in `cost_fpath` file.
            This layer defines the multipliers applied to the cost layer
            to determine LCP routing (but does not actually affect
            output costs). By default, :obj:`BARRIER_H5_LAYER_NAME`.
        barrier_mult : int, optional
            Multiplier on transmission barrier costs.
            By default, ``100``.
        save_paths : bool, optional
            Flag to save reinforcement line path as a multi-line
            geometry. By default, ``False``.
        length_invariant_cost_layers : List[str] | None, optional
            List of layers in H5 to be added to the 'base' greenfield
            cost raster. The costs specified by these layers are not
            scaled with distance traversed across the cell (i.e. fixed
            one-time costs for crossing these cells).
        tracked_layers : dict, optional
            Dictionary mapping layer names to strings, where the strings
            are numpy methods that should be applied to the layer along
            the LCP. For example,
            ``tracked_layers={'layer_1': 'mean', 'layer_2': 'max}``
            would report the average of ``layer_1`` values along the
            least cost path and the max of ``layer_2`` values along the
            least cost path. Examples of numpy methods (non-exhaustive):

                - mean
                - max
                - min
                - mode
                - median
                - std

            By default, ``None``, which does not track any extra layers.
        cell_size : int, optional
            Side length of each cell, in meters. Cells are assumed to be
            square. By default, :obj:`CELL_SIZE`.

        Returns
        -------
        tie_lines : pandas.DataFrame | gpd.GeoDataFrame
            DataFrame of lengths and costs for each reinforcement line
            path or GeoDataFrame of length, cost, and geometry for each
            reinforcement line path.
        """
        ts = time.time()
        tlc = cls(transmission_lines, cost_fpath, start_indices,
                  capacity_class, cost_layers, row_slice, col_slice,
                  xmission_config=xmission_config,
                  tb_layer_name=tb_layer_name, barrier_mult=barrier_mult,
                  length_invariant_cost_layers=length_invariant_cost_layers,
                  tracked_layers=tracked_layers, cell_size=cell_size)

        tie_lines = tlc.compute(end_indices, save_paths=save_paths)
        tie_lines['cost'] = tie_lines['cost'] * 0.5

        row, col = start_indices
        with ExclusionLayers(cost_fpath) as f:
            tie_lines['poi_lat'] = (
                f['latitude', row_slice, col_slice][row, col])
            tie_lines['poi_lon'] = (
                f['longitude', row_slice, col_slice][row, col])

        tie_lines = tie_lines.rename({'length_km': 'reinforcement_dist_km',
                                      'cost': 'reinforcement_cost_per_mw',
                                      'poi_lat': 'reinforcement_poi_lat',
                                      'poi_lon': 'reinforcement_poi_lon'},
                                     axis=1)

        logger.debug('Reinforcement Path Cost computed in %.4f min',
                     (time.time() - ts) / 60)
        logger.debug('Reinforcement tie-line columns: %s',
                     str(tie_lines.columns))

        return tie_lines


def _compute_individual_layers_costs_lens(layer_map, indices, lens, results,
                                          scale_by_length=True,
                                          cell_size=CELL_SIZE):
    """Compute costs and dists by layer.

    Parameters
    ----------
    layer_map : dict
        Dictionary mapping layer names to layer costs. The costs through
        each layer will be computed and reported separately.
    indices : array-like
        Array of (x, y) indices corresponding to the LCP route.
    lens : array-like
        Array of lengths that the route takes through every pixel.
    results : dict
        Dictionary of results for this path to which layer costs will be
        added.
    scale_by_length : bool, optional
        Option to scale costs by length. By default, ``True``.
    cell_size : int, optional
        Side length of each cell, in meters. Cells are assumed to be
        square. By default, :obj:`CELL_SIZE`.

    Returns
    -------
    dict
        Dictionary of results for this path with individual layer costs
        and distances added.
    """
    for layer_name, layer_cost_arr in layer_map.items():
        layer_costs = layer_cost_arr[indices[:, 0], indices[:, 1]]
        if scale_by_length:
            layer_cost = np.sum(layer_costs * lens)
        else:
            layer_cost = np.sum(layer_costs)
        results[f'{layer_name}_cost'] = layer_cost.astype(float)

        # Get path length in km only where layer costs are > 0
        layer_length = np.sum(lens[layer_costs > 0]) * cell_size / 1000
        results[f'{layer_name}_dist_km'] = layer_length.astype(float)

    return results


def _compute_length_mult(features, kind="linear"):
    """Compute length mults based on user input.

    Length multiplier data source:
    https://www.wecc.org/Administrative/TEPPC_TransCapCostCalculator_E3_2019_Update.xlsx
    """
    if kind.casefold() == "step":
        return _compute_step_wise_lm(features)

    if kind.casefold() == "linear":
        return _compute_linear_lm(features)

    raise ValueError(f"Unknown length computation kind: {kind}")


def _compute_step_wise_lm(features):
    """Compute length mults using step function.

    This was the _intented_ original implementation, though the first
    pass was bugged and not fixed before the switch the linear mults
    was made.

    The implemetnation below works as originally intended.
    """
    # Length multiplier
    features['length_mult'] = 1.0
    # Medium cutoff
    mask = features['dist_km'] <= MEDIUM_CUTOFF
    features.loc[mask, 'length_mult'] = MEDIUM_MULT
    # Short cutoff
    mask = features['dist_km'] < SHORT_CUTOFF
    features.loc[mask, 'length_mult'] = SHORT_MULT
    return features


def _compute_linear_lm(features):
    """Compute length mults using linear interpolatiuon below 10 miles."""

    # Length multiplier
    features['length_mult'] = 1.0
    slope = (1 - SHORT_MULT) / (MEDIUM_CUTOFF - SHORT_CUTOFF / 2)

    mask = features['dist_km'] <= MEDIUM_CUTOFF
    features.loc[mask, 'length_mult'] = (
        slope * (features.loc[mask, 'dist_km'] - MEDIUM_CUTOFF) + 1
    )

    return features



# -*- coding: utf-8 -*-
"""
reVX handler utilities.
"""

from .database import Database
from .geopackage import GPKGMeta
from .geotiff import Geotiff
from .layered_h5 import LayeredH5, LayeredTransmissionH5
from .outputs import Outputs
from .sc_points import SupplyCurvePoints



# -*- coding: utf-8 -*-
"""
Setbacks CLI
"""
import os
import glob
import shutil
import logging
from pathlib import Path
from warnings import warn

from gaps.config import load_config
from gaps.cli import CLICommandFromFunction, make_cli
from gaps.cli.preprocessing import preprocess_collect_config
from reVX.setbacks import SETBACKS
from reVX.setbacks.regulations import (validate_setback_regulations_input,
                                       select_setback_regulations)
from reVX.setbacks.setbacks_converter import parse_setbacks
from reVX.handlers.geotiff import Geotiff
from reVX import __version__
from reVX.setbacks.setbacks import SETBACK_SPECS, setbacks_calculator


logger = logging.getLogger(__name__)


def _validate_multipliers(features, generic_setback_multiplier=None):
    """Load, format, and validate generic multipliers. """
    if isinstance(generic_setback_multiplier, str):
        logger.debug("Loading multipliers from {!r}"
                     .format(generic_setback_multiplier))
        generic_setback_multiplier = load_config(generic_setback_multiplier)

    try:
        missing = {feat for feat in features
                   if feat not in generic_setback_multiplier}
    except TypeError:
        return {feat: generic_setback_multiplier for feat in features}

    if missing:
        msg = ("The following features are missing from the "
               "`generic_cost_multiplier` input: {}. Please make sure every "
               "key in the `features` input has a matching key in the "
               "`generic_cost_multiplier` dictionary"
               .format(missing))
        logger.error(msg)
        raise KeyError(msg)
    return {feat: generic_setback_multiplier[feat] for feat in features}


def preprocess_setbacks_config(config, features,
                               generic_setback_multiplier=None):
    """Pre-process setbacks computation config.

    Parameters
    ----------
    config : dict
        Setbacks compute config. This config will be updated to include
        a the ``node_feature_type``, ``node_file_path``, and
        ``node_multiplier`` keys based on user input.
    features : dict
        Dictionary specifying which features/data to process. The keys
        of this dictionary must be the a key from the
        :attr:`~reVX.setbacks.setbacks.SETBACK_SPECS` dictionary or the
        ``feature_specs`` input dictionary specifying the feature type
        to run setbacks for. The value of each key must be a path or a
        list of paths to calculate that particular setback for.
        The path(s) can contain unix-style file-pattern matching syntax
        to point to multiple files. The paths may be specified relative
        to the config file. For example::

            features: {
                "parcel": "../relative/path/to/parcel_colorado.gpkg",
                "road": [
                    "/full/path/to/road/data/*.gpkg",
                    "../../relative/path/to/data_i[l,n].gpkg",
                ]
            }

        With this input, parcel setbacks would be computed for the
        data in ``../relative/path/to/parcel_colorado.gpkg``, and road
        setbacks would be calculated for *all* GeoPackage data files in
        ``/full/path/to/road/data/`` and for the files
        ``../../relative/path/to/data_il.gpkg`` and
        ``../../relative/path/to/data_in.gpkg``.
    generic_setback_multiplier : int | float | str, optional
        Optional setback multiplier to use where local regulations are
        not supplied. This multiplier will be applied to the
        ``base_setback_dist`` (or the turbine tip-height) to calculate
        the setback. If supplied along with ``regulations_fpath``, this
        input will be used to apply a setback to all counties not listed
        in the regulations file. This input can also be a path to a
        config file containing feature types as keys and
        feature-specific generic multipliers as values. For example::

            {
                "parcel": 1.1,
                "road": 2,
                "structure": 3.5
            }

        If specified this way, every key in the ``features`` inputs
        must also be given in the generic multipliers config. If
        ``None``, no  generic setback computation is performed.
        By default, ``None``.


    Returns
    -------
    config : dict
        Updated setbacks computation config dictionary.

    Raises
    ------
    ValueError
        If ``features`` is not a dictionary, or any key in ``features``
        is not a valid key of :attr:`SETBACKS`.
    FileNotFoundError
        If no input GeoPackage files are found on disk.
    """
    if not isinstance(features, dict):
        raise ValueError("`features` key must be a dictionary, got {}"
                         .format(features))
    feature_specs = config.get("feature_specs", {})
    combos_to_run = []
    multipliers = _validate_multipliers(features, generic_setback_multiplier)
    for feature_type, features_fpath in features.items():
        if feature_type not in set(SETBACKS) | set(feature_specs):
            msg = ("feature_type must be one of: {}; got {!r}"
                   .format(set(SETBACKS.keys()), feature_type))
            raise ValueError(msg)

        if isinstance(features_fpath, str):
            features_fpath = [features_fpath]

        for path in features_fpath:
            glob_path = Path(path).expanduser().resolve()
            if glob_path.is_dir():
                glob_path = glob_path / '*'

            need_to_run = [(feature_type, str(f), multipliers[feature_type])
                           for f in glob_path.parent.glob(glob_path.name)
                           if f.name.endswith("gpkg")]

            if not need_to_run:
                msg = ("Found no unprocessed GeoPackage files matching the "
                       "input {!r}!".format(path))
                logger.warning(msg)
                warn(msg)
                continue

            combos_to_run += need_to_run

    if not combos_to_run:
        msg = 'No unprocessed GeoPackage files found!'
        logger.error(msg)
        raise FileNotFoundError(msg)

    feature_type, file_path, multiplier = zip(*sorted(combos_to_run))
    config["node_feature_type"] = feature_type
    config["node_file_path"] = file_path
    config["node_multiplier"] = multiplier
    validate_setback_regulations_input(config.get("base_setback_dist"),
                                       config.get("hub_height"),
                                       config.get("rotor_diameter"))
    _update_setbacks_calculators(feature_specs)  # test for errors
    return config


def preprocess_merge_config(config, project_dir, command_name,
                            merge_file_pattern="PIPELINE"):
    """Pre-process merge config.

    Parameters
    ----------
    config : dict
        Collection config. This config will be updated to include the
        keys ``node_out_path`` and ``node_pattern`` representing the
        output file path and the input file pattern, respectively.
    project_dir : path-like
        Path to project directory. This path is used to resolve the
        output filepath.
    command_name : str
        Name of the command being run. This is used to parse the
        pipeline status for output files if
        ``"merge_file_pattern": "PIPELINE"`` in the input `config`.
    merge_file_pattern : str | list | dict, optional
        Unix-style ``/filepath/pattern*.h5`` representing the files to
        be merged into a single output GeoTIFF file. If no output file
        path is specified (i.e. this input is a single pattern or a list
        of patterns), the output file path will be inferred from the
        pattern itself (specifically, the wildcard will be removed
        and the result will be the output file path). If a list of
        patterns is provided, each pattern will be merged into a
        separate output file. To specify the name of the output file(s),
        set this input to a dictionary whose keys are paths to the
        output file (relative paths are allowed) and the values are
        patterns representing the input files that should be merged into
        the output TIFF. If running a merge job as part of a pipeline,
        this input can be set to ``"PIPELINE"``, which will parse the
        output of the previous step (``compute``) and generate the input
        file pattern and output file name automatically.
        By default, ``"PIPELINE"``.

    Returns
    -------
    config : dict
        Updated merge config dictionary.
    """
    config = preprocess_collect_config(config, project_dir, command_name,
                                       collect_pattern=merge_file_pattern)
    config["node_out_path"] = config.pop("_out_path", None)
    config["node_pattern"] = config.pop("_pattern", None)
    return config


def _update_setbacks_calculators(feature_specs=None):
    """Update `SETBACKS` to include new specs """

    for feature_name, feature_spec in (feature_specs or {}).items():
        spec = SETBACK_SPECS.get(feature_name, {})
        spec.update(feature_spec)
        SETBACKS[feature_name] = setbacks_calculator(**spec)


def compute_setbacks(excl_fpath, node_feature_type, node_file_path,
                     node_multiplier, out_dir, tag, hub_height=None,
                     rotor_diameter=None, base_setback_dist=None,
                     regulations_fpath=None,
                     weights_calculation_upscale_factor=None,
                     replace=False, hsds=False, out_layers=None,
                     feature_specs=None, max_workers=None):
    """Compute Setbacks.

    Setbacks can be computed for a specific turbine (hub height and
    rotor diameter) or more generally using a base setback distance.

    Setbacks can be computed either locally (on a per-county basis with
    given distances/multipliers) or everywhere under a generic setback
    multiplier assumption applied to either the turbine tip-height or
    the base setback distance. These two methods can also be applied
    simultaneously - local setbacks are computed where given (via a the
    regulation file input) and a generic multiplier applied to the
    turbine tip-height or the base setback distance everywhere else.

    Partial inclusions can be computed instead of boolean exclusions,
    both of which can be fed directly into ``reV``.

    Parameters
    ----------
    excl_fpath : str
        Path to HDF5 file containing the county FIPS layer (should be
        called ``cnty_fips``) used to match local regulations in
        ``regulations_fpath`` to counties on the grid. No data will be
        written to this file unless explicitly requested via the
        ``out_layers`` input.
    node_feature_type : str
        Name of the feature type being run. Must be a key of the
        :attr:`SETBACKS` dictionary.
    node_file_path : str
        Path to input feature file. This file MUST be a GeoPackage (and
        have the ".gpkg" extension).
    node_multiplier : int | float | str | None, optional
        A setback multiplier to use if regulations are not supplied.
        This multiplier will be applied to the ``base_setback_dist``
        to calculate the setback. If supplied along with
        ``regulations_fpath``, this input will be used to apply a
        setback to all counties not listed in the regulations file.
    out_dir : str
        Path to output directory where output file should be written.
    tag : str
        Tag to add to each output file to make it unique (i.e. not clash
        with output files from other nodes).
    hub_height : int | float, optional
        Turbine hub height (m), used along with rotor diameter to
        compute the blade tip-height which is used as the base setback
        distance for generic/local regulations. If this input is
        specified, ``rotor_diameter`` must also be given, and
        ``base_setback_dist`` *must be set to None*, otherwise an
        error in thrown. The base setback distance is scaled by
        generic/local multipliers (provided either via the
        ``regulations_fpath`` csv, or the ``generic_setback_multiplier``
        input, or both) before setbacks are computed.
        By default, ``None``.
    rotor_diameter : int | float, optional
        Turbine rotor diameter (m), used along with hub height to
        compute the blade tip-height, which is used as the base setback
        distance for generic/local regulations. If this input is
        specified, ``hub_height`` must also be given, and
        ``base_setback_dist`` *must be set to None*, otherwise an
        error in thrown. The base setback distance is scaled by
        generic/local multipliers (provided either via the
        ``regulations_fpath`` csv, or the ``generic_setback_multiplier``
        input, or both) before setbacks are computed.
        By default, ``None``.
    base_setback_dist : int | float, optional
        Base setback distance (m). This value is used as the base
        setback distance for generic/local regulations. If this input is
        specified, both ``hub_height``and ``rotor_diameter`` *must be
        set to None*, otherwise an error in thrown. The base setback
        distance is scaled by generic/local multipliers (provided either
        via the ``regulations_fpath`` csv, or the
        ``generic_setback_multiplier`` input, or both) before setbacks
        are computed. By default, ``None``.
    regulations_fpath : str, optional
        Path to regulations ``.csv`` or ``.gpkg`` file. At a minimum,
        this file must contain the following columns:

            - ``Feature Type``: Contains labels for the type of setback
              that each row represents. This should be a
              `"feature_type"` label that can be found in the
              :attr:`~reVX.setbacks.setbacks.SETBACK_SPECS` dictionary
              (e.g. ``"structures"``, ``"roads"``, ``"water"``, etc.),
              unless you have created your own setback calculator using
              :func:`~reVX.setbacks.setbacks.setbacks_calculator`, in
              which case this label can match the `feature_type` input
              you used for that function call.
            - ``Feature Subtype``: Contains labels for feature subtypes.
              The feature subtypes are only used for down-selecting the
              local regulations that should be applied for a particular
              feature, so often you can leave this blank or set it to
              ``None``. If you do specify this value, it should be a
              `"feature_subtypes_to_exclude"` label that can be found in
              the
              :attr:`~reVX.setbacks.setbacks.SETBACK_SPECS` dictionary,
              unless you have created your own setback calculator using
              :func:`~reVX.setbacks.setbacks.setbacks_calculator`, in
              which case this label can match the
              `feature_subtypes_to_exclude` input you used for that
              function call.
            - ``Value Type``: Specifies wether the value is a multiplier
              or static height.  See
              :obj:`~reVX.setbacks.regulations.SetbackRegulations` (if
              using only ``base_setback_dist`` input) or
              :obj:`~reVX.setbacks.regulations.WindSetbackRegulations`
              (if using ``hub_height`` + ``rotor_diameter`` input) for
              more info.
            - ``Value``: Numeric value of the setback or multiplier.
            - ``FIPS``: Specifies a unique 5-digit code for each county
              (this can be an integer - no leading zeros required). This
              is used along side the ``cnty_fips`` layer in the
              `excl_fpath` to match the county regulations to the
              county's spatial extent.

        This option overrides the ``generic_setback_multiplier`` input,
        but only for counties that are listed in the input CSV file.
        This means both ``regulations_fpath`` and
        ``generic_setback_multiplier`` can be specified
        simultaneously in order to compute setbacks driven by local
        ordinance where given + a generic multiplier applied everywhere
        else. By default, ``None``, which does not compute any
        local setbacks.
    weights_calculation_upscale_factor : int, optional
        Optional input to specify *partial* setback calculations.
        If this value is an int > 1, the output will be a layer with
        **inclusion** weight values (floats ranging from 0 to 1). Note
        that this is backwards w.r.t the typical output of exclusion
        integer values (1 for excluded, 0 otherwise). Values <= 1 will
        still return a standard exclusion mask. For example, a cell that
        was previously excluded with a boolean mask (value of 1) may
        instead be converted to an inclusion weight value of 0.75,
        meaning that 75% of the area corresponding to that point should
        be included (i.e. the exclusion feature only intersected a small
        portion - 25% - of the cell). This percentage inclusion value is
        calculated by upscaling the output array using this input value,
        rasterizing the exclusion features onto it, and counting the
        number of resulting sub-cells excluded by the feature. For
        example, setting the value to ``3`` would split each output cell
        into nine sub-cells - 3 divisions in each dimension. After the
        feature is rasterized on this high-resolution sub-grid, the area
        of the non-excluded sub-cells is totaled and divided by the area
        of the original cell to obtain the final inclusion percentage.
        Therefore, a larger upscale factor results in more accurate
        percentage values. If ``None`` (or a value <= 1), this process
        is skipped and the output is a boolean exclusion mask.
        By default ``None``.
    replace : bool, optional
        Flag to replace the output GeoTIFF if it already exists.
        By default, ``False``.
    hsds : bool, optional
        Boolean flag to use ``h5pyd`` to handle HDF5 "files" hosted on
        AWS behind HSDS. By default, ``False``.
    out_layers : dict, optional
        Dictionary mapping the input feature file names (with extension)
        to names of layers under which exclusions should be saved in the
        ``excl_fpath`` HDF5 file. If ``None`` or empty dictionary,
        no layers are saved to the HDF5 file. By default, ``None``.
    feature_specs : dict, optional
        Optional dictionary specifying new feature setback calculators
        or updates to existing ones. The keys of this dictionary should
        be names of the features for which a specification is being
        provided. If the name is already a key in
        :attr:`~reVX.setbacks.setbacks.SETBACK_SPECS`, the corresponding
        specifications wil be updated for that feature. Otherwise, the
        name will represent a new feature type, which can be used as a
        key in the ``features`` input. The values of the feature-type
        keys should be dictionaries, where the keys are parameters of
        the :func:`~reVX.setbacks.setbacks.setbacks_calculator`
        function. Required parameters in that function are required keys
        of these dictionaries. Values should be the updated value.
        For example, the input
        ::

            feature_specs: {
                "water": {
                    "num_features_per_worker": 500
                },
                "oil_and_gas_pipelines": {
                    "feature_type": "oil and gas",
                    "feature_filter_type": "clip"
                }
            }

        would update the existing ``"water"`` setbacks calculator to
        compute 500 features per worker at a time and create a new
        ``"oil_and_gas_pipelines"`` feature that looks for the string
        ``"oil and gas"`` in the regulations file and clips the feature
        to a county before calculating a setback. Note that even though
        ``"oil_and_gas_pipelines"`` is not a default feature supported
        by ``reVX``, you can now use it in the ``features`` input.
        This can also be helpful if you need to compute the same type of
        setback for multiple different input datasets. For example, the
        input
        ::

            feature_specs: {
                "water-nwi": {
                    "feature_type": "water",
                    "buffer_type": "default",
                    "feature_filter_type": "clip",
                    "num_features_per_worker": 700,
                },
                "water-nhd": {
                    "feature_type": "water",
                    "buffer_type": "default",
                    "feature_filter_type": "clip",
                    "num_features_per_worker": 10_000,
                }
            }

        would allow you to set up your ``features`` input like so::

            features: {
                "water-nwi": "/path/to/nwi/*.gpkg",
                "water-nhd": "/path/to/nhd/*.gpkg",
            }

        By default, ``None``, which does not add any new setback
        calculators (the default ones defined in
        :attr:`~reVX.setbacks.setbacks.SETBACK_SPECS` are still
        available).
    max_workers : int, optional
        Number of workers to use for setback exclusion computation. If
        this value is 1, the computation runs in serial. If this value
        is > 1, the computation runs in parallel with that many workers.
        If ``None``, the computation runs in parallel on all available
        cores. By default, ``None``.

    Returns
    -------
    str
        Path to output GeoTIFF file containing setback exclusion data.
    """

    _update_setbacks_calculators(feature_specs)
    logger.info('Computing setbacks from {} in {}'
                .format(node_feature_type, node_file_path))
    logger.debug('Setbacks to be computed with:\n'
                 '- base_setback_dist = {}\n'
                 '- hub_height = {}\n'
                 '- rotor_diameter = {}\n'
                 '- regulations_fpath = {}\n'
                 '- generic_setback_multiplier = {}\n'
                 '- using max_workers = {}\n'
                 '- replace layer if needed = {}\n'
                 '- weights calculation upscale factor = {}\n'
                 '- out_layers = {}\n'
                 .format(base_setback_dist, hub_height, rotor_diameter,
                         regulations_fpath, node_multiplier, max_workers,
                         replace,
                         weights_calculation_upscale_factor, out_layers))

    regulations = select_setback_regulations(base_setback_dist, hub_height,
                                             rotor_diameter, regulations_fpath,
                                             node_multiplier)
    setbacks_class = SETBACKS[node_feature_type]
    wcuf = weights_calculation_upscale_factor
    fn = ("setbacks_{}_{}{}.tif"
          .format(node_feature_type, os.path.basename(out_dir), tag))
    out_fn = os.path.join(out_dir, fn)
    setbacks_class.run(excl_fpath, node_file_path, out_fn, regulations,
                       weights_calculation_upscale_factor=wcuf,
                       max_workers=max_workers, replace=replace, hsds=hsds,
                       out_layers=out_layers)
    logger.info('Setbacks computed and written to {}'.format(out_fn))
    return out_fn


def merge_setbacks(node_out_path, node_pattern, are_partial_inclusions=None,
                   purge_chunks=False):
    """Combine many input setback GeoTIFFs into a single layer.

    Parameters
    ----------
    node_out_path : str
        Path to output GeoTIFF file.
    node_pattern : str
        Input GeoTIFF file pattern.
    are_partial_inclusions : bool, optional
        Flag indicating wether the inputs are partial inclusion values
        or boolean exclusions. If ``None``, will try to infer
        automatically from the input file's GeoTIFF profile
        (``dtype != uint8``). By default, ``None``.
    purge_chunks : bool, optional
        Flag indicating wether individual "chunk" files should be
        deleted after a successful merge (``True``), or if they should
        be stored in a "chunk_files" directory (``False``).
        By default, ``False``.

    Raises
    ------
    FileNotFoundError
        If no input files found on disk matching the input pattern.
    RuntimeError
        If the ``are_partial_inclusions`` cannot be inferred (GeoTIFF
        profile does not have "dtype" field).
    """
    out_file = Path(node_out_path).resolve()

    logger.info("Merging TIFF files in {!r} and writing to {!r}"
                .format(out_file.parent.as_posix(), out_file))
    input_setback_files = list(glob.glob(node_pattern))
    if not input_setback_files:
        msg = ("Did not find any files matching pattern {!r} in directory {!r}"
               .format(node_pattern, out_file.parent.as_posix()))
        logger.error(msg)
        raise FileNotFoundError(msg)

    with Geotiff(input_setback_files[0]) as tif:
        profile = tif.profile

    if are_partial_inclusions is None:
        data_dtype = profile.get("dtype")
        if data_dtype is None:
            msg = ("Cannot infer whether data are partial inclusions because "
                   "data dtype is not set in the GeoTIFF profile. Please "
                   "explicitly specify the `are_partial_inclusions` input!")
            logger.error(msg)
            raise RuntimeError(msg)

        are_partial_inclusions = data_dtype.casefold() != "uint8"

    setbacks = parse_setbacks(input_setback_files,
                              is_inclusion_layer=are_partial_inclusions)

    logger.info("Writing data to {!r}".format(out_file.as_posix()))
    Geotiff.write(out_file.as_posix(), profile, setbacks)

    if purge_chunks:
        for fpath in input_setback_files:
            os.remove(fpath)
    else:
        chunk_dir = out_file.parent / "chunk_files"
        chunk_dir.mkdir(parents=True, exist_ok=True)
        for fpath in input_setback_files:
            shutil.move(fpath, chunk_dir / os.path.basename(fpath))

        logger.info("Moved chunk files from {} to sub_dir: {}"
                    .format(out_file.parent, chunk_dir))


PRIVATE_COMPUTE_KEYS = ("node_feature_type", "node_file_path",
                        "node_multiplier")
PRIVATE_MERGE_KEYS = ("node_out_path", "node_pattern")
commands = [
    CLICommandFromFunction(
        function=compute_setbacks, name="compute",
        split_keys=[PRIVATE_COMPUTE_KEYS],
        config_preprocessor=preprocess_setbacks_config,
        skip_doc_params=PRIVATE_COMPUTE_KEYS,
    ),
    CLICommandFromFunction(
        function=merge_setbacks, name="merge",
        split_keys=[PRIVATE_MERGE_KEYS],
        config_preprocessor=preprocess_merge_config,
        skip_doc_params=PRIVATE_MERGE_KEYS,
    ),
]

cli = make_cli(commands)


if __name__ == '__main__':
    try:
        cli(obj={})
    except Exception:
        logger.exception('Error running Setbacks CLI')
        raise



"""
reVX rasterization utilities
"""
import logging

from rasterio import features

logger = logging.getLogger(__name__)


def rasterize(gdf, profile, buffer_dist=None, all_touched=False, burn_value=1,
              boundary_only=False, dtype="float32"):
    """
    Rasterize a vector layer.

    Parameters
    ----------
    gdf : geopandas.DataFrame
        Geopandas DataFrame containins shapes to rasterize.
    profile : Profile
        Raster profile to use.
    buffer_dist : float, optional
        Distance to buffer features in fname by. Same units as the
        template raster. By default, ``None``.
    all_touched : bool, optional
        Set all cells touched by vector to 1. False results in less cells
        being set to 1. By default, ``False``.
    burn_value : int | float, optional
        Value used to burn vectors into raster. By default, ``1``.
    boundary_only : bool, optional
        If True, rasterize boundary of vector. By default, ``False``.
    dtype : np.dtype, optional
        Datatype to use. By default, ``float32``.

    Returns
    -------
    numpy.nd_array
        Rasterized vector data
    """

    if buffer_dist is not None:
        logger.debug('Buffering shapes by %s', buffer_dist)
        gdf.geometry = gdf.geometry.buffer(buffer_dist)
        logger.debug('Buffering done. %d features before cleaning.', len(gdf))
        gdf = gdf[~gdf.is_empty]  # Negative buffer may result in empty feats
        logger.debug('%d features after removing empty features.', len(gdf))

    logger.debug('Rasterizing shapes')
    geo = gdf.boundary if boundary_only else gdf.geometry
    shape = (profile['height'], profile['width'])
    rasterized = features.rasterize(list(geo), out_shape=shape,
                                    fill=0, out=None,
                                    transform=profile['transform'],
                                    all_touched=all_touched,
                                    default_value=burn_value, dtype=dtype)

    logger.debug('Rasterizing complete')
    return rasterized



"""
Create, load, and store masks to determine land and sea.
"""
import os
import logging
from pathlib import Path
from typing import Optional

import numpy as np
import numpy.typing as npt

from reVX.handlers.layered_h5 import LayeredTransmissionH5
from reVX.least_cost_xmission.layers.utils import rasterize_shape_file

logger = logging.getLogger(__name__)

# Mask array
MaskArr = npt.NDArray[np.bool_]

MASK_MSG = ('No mask available. Please run create_masks() or load_masks() '
            'first.')


class Masks:
    """
    Create, load, and store masks to determine land and sea.
    """
    LANDFALL_MASK_FNAME = 'landfall_mask.tif'
    """One pixel width line at shore"""
    RAW_LAND_MASK_FNAME = 'raw_land_mask.tif'
    """Rasterized land vector"""
    LAND_MASK_FNAME = 'land_mask.tif'
    """Raw mask - landfall mask"""
    OFFSHORE_MASK_FNAME = 'offshore_mask.tif'
    """Offshore mask filename"""

    def __init__(self, io_handler: LayeredTransmissionH5, masks_dir='.'):
        """

        Parameters
        ----------
        io_handler : :class:`LayeredTransmissionH5`
            Transmission IO handler
        masks_dir : path-like, optional
            Directory for storing/finding mask GeoTIFFs. By default, ``'.'``.
        """
        self._io_handler = io_handler
        self._masks_dir = masks_dir
        os.makedirs(masks_dir, exist_ok=True)

        self._landfall_mask: Optional[MaskArr] = None
        self._dry_mask: Optional[MaskArr] = None
        self._wet_mask: Optional[MaskArr] = None
        self._dry_plus_mask: Optional[MaskArr] = None
        self._wet_plus_mask: Optional[MaskArr] = None

    @property
    def landfall_mask(self) -> MaskArr:
        """MaskArr: Landfalls cells mask, only one cell wide """
        if self._landfall_mask is None:
            raise ValueError(MASK_MSG)
        return self._landfall_mask

    @property
    def wet_mask(self) -> MaskArr:
        """MaskArr: Wet cells mask, does not include landfall cells """
        if self._wet_mask is None:
            raise ValueError(MASK_MSG)
        return self._wet_mask

    @property
    def dry_mask(self) -> MaskArr:
        """MaskArr: Dry cells mask, does not include landfall cells """
        if self._dry_mask is None:
            raise ValueError(MASK_MSG)
        return self._dry_mask

    @property
    def dry_plus_mask(self) -> MaskArr:
        """MaskArr: Dry cells mask, includes landfall cells """
        if self._dry_plus_mask is None:
            self._dry_plus_mask = np.logical_or(self.dry_mask,
                                                self.landfall_mask)
        return self._dry_plus_mask

    @property
    def wet_plus_mask(self) -> MaskArr:
        """MaskArr: Wet cells mask, includes landfall cells """
        if self._wet_plus_mask is None:
            self._wet_plus_mask = np.logical_or(self.wet_mask,
                                                self.landfall_mask)
        return self._wet_plus_mask

    def create_masks(self, land_mask_shp_f: str, save_tiff: bool = True,
                     reproject_vector: bool = True):
        """
        Create the offshore and land mask layers from a polygon land vector
        file.

        Parameters
        ----------
        mask_shp_f : str
            Full path to land polygon gpgk or shp file
        save_tiff : bool, optional
            Save mask as tiff if true. By default, ``True``.
        reproject_vector : bool, optional
            Reproject CRS of vector to match template raster if True.
            By default, ``True``.
        """
        logger.debug('Creating masks from %s', land_mask_shp_f)

        # Raw land is all land cells, include landfall cells
        raw_land = rasterize_shape_file(land_mask_shp_f,
                                        self._io_handler.profile,
                                        all_touched=True,
                                        reproject_vector=reproject_vector,
                                        dtype='uint8')

        raw_land_mask: MaskArr = raw_land == 1

        # Offshore mask is inversion of raw land mask
        self._wet_mask = ~raw_land_mask

        landfall = rasterize_shape_file(land_mask_shp_f,
                                        self._io_handler.profile,
                                        reproject_vector=reproject_vector,
                                        all_touched=True, boundary_only=True,
                                        dtype='uint8')
        self._landfall_mask = landfall == 1

        # XOR landfall and raw land to get all land cells, except landfall
        # cells
        self._dry_mask = np.logical_xor(self.landfall_mask,
                                        raw_land_mask)

        logger.debug('Created all masks')

        if save_tiff:
            logger.debug('Saving masks to GeoTIFF')
            self._save_mask(raw_land_mask, self.RAW_LAND_MASK_FNAME)
            self._save_mask(self.wet_mask, self.OFFSHORE_MASK_FNAME)
            self._save_mask(self.dry_mask, self.LAND_MASK_FNAME)
            self._save_mask(self.landfall_mask, self.LANDFALL_MASK_FNAME)
            logger.debug('Completed saving all masks')

    def load_masks(self):
        """
        Load the mask layers from GeoTIFFs. This does not need to be called if
        self.create_masks() was run previously. Mask files must be in the
        current directory.
        """
        logger.debug('Loading masks')
        self._dry_mask = self._load_mask(self.LAND_MASK_FNAME)
        self._wet_mask = self._load_mask(self.OFFSHORE_MASK_FNAME)
        self._landfall_mask = self._load_mask(self.LANDFALL_MASK_FNAME)
        logger.debug('Successfully loaded wet, dry, and landfall masks')

    def _save_mask(self, data: npt.NDArray, fname: str):
        """
        Save mask to GeoTiff

        Parameters
        ----------
        data
            Data to save in GeoTiff
        fname
            Name of file to save
        """
        full_fname = os.path.join(self._masks_dir, fname)
        self._io_handler.save_data_using_h5_profile(data, full_fname)

    def _load_mask(self, fname: str) -> npt.NDArray[np.bool_]:
        """
        Load mask from GeoTIFF with sanity checking

        Parameters
        ----------
        fname
            Filename to load mask from

        Returns
        -------
            Mask data
        """
        full_fname = os.path.join(self._masks_dir, fname)

        if not Path(full_fname).exists():
            raise FileNotFoundError(f'Mask file at {full_fname} not found. '
                                    'Please create masks first.')

        raster = self._io_handler.load_data_using_h5_profile(full_fname,
                                                             reproject=True)

        if raster.max() != 1:
            msg = (f'Maximum value in mask file {fname} is {raster.max()} but'
                   ' should be 1. Mask file appears to be corrupt. Please '
                   'recreate it.')
            logger.error(msg)
            raise ValueError(msg)
        if raster.min() != 0:
            msg = (f'Minimum value in mask file {fname} is {raster.min()} but'
                   ' should be 0. Mask file appears to be corrupt. Please '
                   'recreate it.')
            logger.error(msg)
            raise ValueError(msg)

        return raster == 1



# -*- coding: utf-8 -*-
"""
Convert dist_to_ports geotiff to .h5 exclusion layers
"""
import logging
import numpy as np
import os
from warnings import warn

from reV.handlers.exclusions import ExclusionLayers

from reVX.handlers.geotiff import Geotiff
from reVX.handlers.layered_h5 import LayeredH5, check_geotiff
from reVX.handlers.outputs import Outputs

logger = logging.getLogger(__name__)


class DistToPortsConverter(LayeredH5):
    """
    Convert Distance to Ports goetiff(s) to excl .h5 layers
    """
    @classmethod
    def _parse_dist_to_ports(cls, dist_to_ports, chunks=(128, 128)):
        """
        Load dist_to_ports, combine multiple dist_to_ports by state if needed

        Parameters
        ----------
        dist_to_ports : list
            List of paths to dist_to_ports geotiffs to load and combine
        chunks : tuple, optional
            Chunk size of exclusions in Geotiff, by default (128, 128)

        Returns
        -------
        values : ndarray
            dist_to_ports exclusion array
        """
        values = None
        for geotiff in dist_to_ports:
            with Geotiff(geotiff, chunks=chunks) as tif:
                v = tif.values
            if not np.any(v):
                msg = ('{} is invalid and only contains zeros and will be '
                       'skipped. It is advised to recreate the file and '
                       'update the distance to ports layer!'
                       .format(geotiff))
                logger.warning(msg)
                warn(msg)
            else:
                if values is None:
                    values = v
                else:
                    values = np.minimum(values, v)

        return values

    def write_dist_to_ports_to_h5(self, dist_to_ports, layer_name,
                                  check_tiff=True, transform_atol=0.01,
                                  description=None, scale_factor=None,
                                  dtype='float32'):
        """
        Transfer geotiff exclusions to h5 confirming they match existing layers

        Parameters
        ----------
        dist_to_ports : str
            Path to directory containing distance to port geotiff files or
            a list of the distance to port geotiff filepaths
        layer_name : str
            Layer to create
        check_tiff : bool, optional
            Flag to check tiff profile and coordinates against exclusion .h5
            profile and coordinates, by default True
        transform_atol : float, optional
            Absolute tolerance parameter when comparing geotiff transform data,
            by default 0.01
        description : str, optional
            Description of exclusion layer, by default None
        scale_factor : int | float, optional
            Scale factor to use to scale geotiff data when added to the .h5
            file, by default None
        dtype : str, optional
            Dtype to save geotiff data as in the .h5 file. Only used when
            'scale_factor' is not None, by default 'float32'
        """
        if os.path.isdir(dist_to_ports):
            dist_to_ports = [os.path.join(dist_to_ports, file)
                             for file in os.listdir(dist_to_ports)
                             if file.endswith('.tif')]
        else:
            dist_to_ports = [dist_to_ports]

        logger.debug('\t- Computing minimum distance to ports from {}'
                     .format(dist_to_ports))

        if not os.path.exists(self.h5_file):
            if self.template_file == self.h5_file:
                self.template_file = dist_to_ports[0]
            self.create_new(overwrite=False)

        if check_tiff:
            check_geotiff(self, dist_to_ports[0], chunks=self._chunks,
                          transform_atol=transform_atol)

        with Geotiff(dist_to_ports[0], chunks=self._chunks) as tif:
            profile = tif.profile

        dist_to_ports = self._parse_dist_to_ports(dist_to_ports,
                                                  chunks=self._chunks)
        if layer_name in self.layers:
            msg = ("{} is already present in {} and will be updated"
                   .format(layer_name, self.h5_file))
            logger.warning(msg)
            warn(msg)
            with ExclusionLayers(self.h5_file) as exc:
                dist_to_ports = np.minimum(dist_to_ports, exc[layer_name])

        if scale_factor is not None:
            attrs = {'scale_factor': scale_factor}
            dist_to_ports = Outputs._check_data_dtype(
                layer_name, dist_to_ports, dtype, attrs=attrs)

        self.write_layer_to_h5(dist_to_ports, layer_name, profile=profile,
                               description=description,
                               scale_factor=scale_factor)

    def layers_to_h5(self, layers, check_tiff=True, transform_atol=0.01,
                     descriptions=None, scale_factors=None):
        """
        Create exclusions .h5 file, or load layers into existing exclusion .h5
        file from provided dist_to_ports

        Parameters
        ----------
        layers : dict
            Dictionary mapping goetiffs to the layers to load
        check_tiff : bool, optional
            Flag to check tiff profile and coordinates against exclusion .h5
            profile and coordinates, by default True
        transform_atol : float, optional
            Absolute tolerance parameter when comparing geotiff transform data,
            by default 0.01
        coord_atol : float, optional
            Absolute tolerance parameter when comparing new un-projected
            geotiff coordinates against previous coordinates, by default 0.001
        description : dict, optional
            Description of exclusion layers, by default None
        scale_factor : dict, optional
            Scale factors and dtypes to use when scaling given layers,
            by default None
        """
        if scale_factors is None:
            scale_factors = {}

        if descriptions is None:
            descriptions = {}

        if isinstance(layers, list):
            layers = {os.path.basename(lyr).split('.')[0]: lyr
                      for lyr in layers}

        logger.info('Moving layers to %s', self.h5_file)
        for layer_name, dist_to_ports in layers.items():
            logger.info('- Transfering %s', layer_name)
            description = descriptions.get(layer_name, None)
            scale = scale_factors.get(layer_name, None)
            if scale is not None:
                scale_factor = scale['scale_factor']
                dtype = scale['dtype']
            else:
                scale_factor = None
                dtype = None

            self.write_dist_to_ports_to_h5(dist_to_ports, layer_name,
                                           check_tiff=check_tiff,
                                           transform_atol=transform_atol,
                                           description=description,
                                           scale_factor=scale_factor,
                                           dtype=dtype)



"""
Definition of friction, barrier, and costs processing JSON config file
"""
from pathlib import Path
from typing import Optional, Dict, List, Literal

from pydantic import BaseModel, DirectoryPath, FilePath

from reVX.least_cost_xmission.config import IsoMultipliers
from reVX.least_cost_xmission.config.constants import (ALL,
                                                       CELL_SIZE,
                                                       BARRIER_H5_LAYER_NAME)


# Terms for specifying masks. 'wet+' and 'dry+' indicated 'wet' + 'landfall'
# and 'dry' + 'landfall', respectively.
Extents = Literal['all', 'wet', 'wet+', 'landfall', 'dry+', 'dry']


class RangeConfig(BaseModel, extra='forbid'):
    """
    Config for defining a range and a value to assign to cells matching that
    range. Cells with values >= than 'min' and < 'max' will be assigned
    'value'. One or both of 'min' and 'max' can be specified.
    """
    min: float = float('-inf')
    """Minimum value to get a cost assigned (inclusive)"""

    max: float = float('inf')
    """Maximum value to get a cost assigned (exclusive)"""

    value: float
    """Value to assign to the range defined by `min` and `max`."""


class Rasterize(BaseModel, extra='forbid'):
    """
    Rasterize a vector layer and apply a value to it.
    """
    value: float
    """Value to burn in to raster"""

    buffer: Optional[float] = None
    """Value to buffer by (can be negative)"""

    reproject: bool = True
    """Reproject vector to raster CRS if ``True``"""


class LayerBuildConfig(BaseModel, extra='forbid'):
    """
    Friction and barrier layers config model.
    'global_value', 'map', 'bins', 'rasterize', and 'forced_inclusion'
    are exclusive, but exactly one must be specified. Example configs
    can be seen in `test_xmission_barrier_friction_builder.py` in the
    tests directory.
    """

    extent: Extents = ALL
    """Extent to apply map or range to.

    Must be one of the following:

        - 'all': Full extent, including offshore, onshore, and landfall
        - 'wet': offshore extent only
        - 'wet+': offshore extent + landfall extent
        - 'landfall': landfall extent (area between wet and dry extents)
        - 'dry+': onshore extent + landfall extent
        - 'dry': onshore extent only

    By default, 'all'.
    """

    global_value: Optional[float] = None
    """Global value to use for entire layer extent. """

    map: Optional[Dict[float, float]] = None
    """Values in raster (keys) and values to use layer."""

    bins: Optional[List[RangeConfig]] = None
    """Ranges of raster values.

    This input can be one or more ranges of raster values to apply to
    barrier/friction. The value of overlapping ranges are added together."""

    pass_through: Optional[bool] = False
    """Pass cost data through without estra processing."""

    rasterize: Optional[Rasterize] = None
    """Rasterize a vector and save as layer"""

    forced_inclusion: bool = False
    """Force inclusion.

    If `forced_inclusion` is ``True``, any cells with a value > 0 will
    force the final value of corresponding cells to 0. Multiple forced
    inclusions are allowed."""


class DryCosts(BaseModel, extra='forbid'):
    """ Config items required to generate dry costs """
    iso_region_tiff: FilePath
    """Filename of ISO region GeoTIFF"""

    nlcd_tiff: FilePath
    """File name of NLCD GeoTiff"""

    slope_tiff: FilePath
    """File name of slope GeoTiff"""

    cost_configs: Optional[FilePath] = None
    """Path to json file with Xmission cost configuration values.

    Path to json file contianing dictionary with Xmission cost
    configuration values. Valid configuration keysare:

        - "base_line_costs"
        - "iso_lookup"
        - "iso_multipliers"
        - "land_use_classes"
        - "new_substation_costs"
        - "power_classes"
        - "power_to_voltage"
        - "transformer_costs"
        - "upgrade_substation_costs"

    Each of these keys should point to a dictionary or a path to
    a separate json file contianing a dictionary of
    configurations for each section."""

    default_mults: Optional[IsoMultipliers] = None
    """Multipliers to be used for default region.

    This input should be a dictionary with three keys:

        - "iso": Thie key is ignored, but is required. Can set to
          "default" and move on.
        - "land_use": A dictionary where keys are the land use types
          (e.g. "cropland", "forest", "wetland", etc.) and values are
          the multipliers for those land uses.
        - "slope": A dictionary where keys are the slope types/mults
          (e.g. "hill_mult", "hill_slope", "mtn_mult", "mtn_slope",
          etc.) and values are the slopes/multipliers.

    """

    extra_tiffs: Optional[List[FilePath]] = None
    """Optional list of extra GeoTIFFs to add to cost H5 file. """


class MergeFrictionBarriers(BaseModel, extra='forbid'):
    """
    Combine friction and barriers and save to H5. Multiple all barrier values
    by a factor. The multiplier should be large enough that all barriers have
    a higher value than any possible friction.
    """

    friction_layer: str
    """Name of friction layer.

    Must be a layer that has just been created or already exists in the
    tiff directory with the same name and '.tif' extension.
    """

    barrier_layer: str
    """Name of barrier layer.

    Must be a layer that has just been created or already exists in the
    tiff directory with the same name and '.tif' extension.
    """

    output_layer_name: Optional[str] = BARRIER_H5_LAYER_NAME
    """Name of combined output layer.

    By default, :obj:`BARRIER_H5_LAYER_NAME`."""

    barrier_multiplier: float = 1e6
    """Value to multiply barrier layer by during merge with friction.

    The multiplier should be large enough that all barriers have
    a higher value than any possible friction."""


LayerBuildComponents = Dict[str, LayerBuildConfig]
"""Mapping of layer components to use for building the final layer.

Keys are GeoTIFF or vector filepaths. Values are the LayerBuildConfig
to use for that file."""


class LayerConfig(BaseModel):
    """
    Definition of friction, barrier, and costs processing JSON config file.
    """

    layer_name: str
    """Name of layer in H5 file. """

    description: Optional[str] = None
    """Optional description to store in attrs for layer. """

    include_in_h5: Optional[bool] = True
    """Flag to specify whether layer should be stored in H5 or not. """

    values_are_costs_per_mile: Optional[bool] = False
    """Option to specify that the values given represent $/mile.

    If ``True``, the values will be converted to $/CELL_DIST, which is
    what is ultimately used for routing.
    """

    build: LayerBuildComponents
    """Mapping of layer components used to build this layer.

    Keys are GeoTIFF or vector filepaths. Values are the
    LayerBuildConfig to use for that file."""


Layers = List[LayerConfig]
"""Layer configs to build and potentially add to H5 file. """


class TransmissionLayerCreationConfig(BaseModel):
    """
    Definition of friction, barrier, and costs processing JSON config file.
    """

    template_raster_fpath: FilePath
    """Template raster GeoTIFF for shape and profile"""

    h5_fpath: FilePath
    """H5 to store results in."""

    layer_dir: DirectoryPath = Path('.')
    """Directory to look for GeoTIFFs in, in addition to '.'"""

    masks_dir: DirectoryPath = Path('.')
    """Optional path for mask GeoTIFFs."""

    output_tiff_dir: DirectoryPath = Path('.')
    """Directory to store output tiff files in. """

    layers: Optional[Layers] = None
    """Optional configuration for layers to be built.

    At least one of `layers`, `dry_costs`, or
    `merge_friction_and_barriers` must be defined.
    """

    dry_costs: Optional[DryCosts] = None
    """Optional dry cost layer.

    At least one of `layers`, `dry_costs`, or
    `merge_friction_and_barriers` must be defined.
    """

    merge_friction_and_barriers: Optional[MergeFrictionBarriers] = None
    """Optional config to merge friction barriers.

    At least one of `layers`, `dry_costs`, or
    `merge_friction_and_barriers` must be defined.
    """

    cell_size: int = CELL_SIZE
    """Side length of each cell, in meters.

    Cells are assumed to be square. By default, :obj:`CELL_SIZE`."""
